/*********************************************************************
 **
 **   server/appx-client-menu.js - Client Menu/Toolbar processing
 **
 **   This module contains code to process Appx Menus and Toolbars.
 **
 *********************************************************************/

// what_str =  "@(#)Appx $Header: /src/cvs/appxHtml5/server/appx-client-menu.js,v 1.48 2015/04/30 20:56:26 pete Exp $";

var menugroups = [];
var menulabels = [];
var menuitems = [];
var menu = {};
var menu2 = {};

function MenuTree() {
    this.fileMenu = null;
    this.userMenu = [];
    this.optnMenu = null;
    this.helpMenu = null;
}

function MenuNode(label, widget) {
    this.label = label;
    this.widget = widget;
    this.children = [];
}

function MenuItem(label, widget) {
    this.label = label;
    this.widget = widget;
}

function mGroup(menuGroup, label, widget) {
    this.menuGroup = menuGroup;
    this.label = label;
    this.widget = widget;
    this.children = [];
}

function mChild(menuLabel, label, widget) {
    this.menuLabel = menuLabel;
    this.label = label;
    this.widget = widget;
}

function initializeMenu() {
    $("#appx-softkeys-container .toolbaritem").remove();
    $("#appx-defaulttools-container .toolbaritem").remove();
    appx_session.currentmenu = {};
    appx_session.currentmenu.groupsprocessed = [];
    appx_session.currentmenu.groups = {};
    appx_session.currentmenu.groupsprocessed.push("top_menu");
    appx_session.menucache = {};
    appx_session.menucache.currentmenuitems = appx_session.currentmenuitems;
    appx_session.menucache.currentmenu = appx_session.currentmenu;
    appx_session.currentmenuitems = {};
    appx_session.currentmenuitems.all = [];
    appx_session.currentmenuitems.toolbar = [
        [],
        [],
        []
    ];
    appx_session.currentmenuitems.dropdown = [
        [],
        [],
        []
    ];
    appx_session.currentmenuitems.popup = [
        [],
        [],
        []
    ];
    appx_session.currentmenu.block = {};
    appx_session.currentmenu.header = [];
}

//MENU Message Handler
function appxmenuhandler(x) {
    appx_session.currentmenu.block = x;
    appx_session.currentmenu.header.push(new Widget(null, null, appx_session.currentmenu.block.data.headerdata));
    //In LOCALOS for testing
    appxmenushandler(x);
}

//Menus Message Handler
function appxmenushandler(x) {
    var menu = x.data;
    var menuType = menu.type[0];
    //split menu.headerdata into vars
    var mw = new Widget(null, null, null);
    mw.widgetData = x.data.headerdata;
    var menuheader = mw.parseDataIntoPairs();
    $.each(menuheader, function (k, v) {
        menuheader[v.key] = v.value;
    });
    //if use=t, reuse last menu
    if (menuheader.USE) {
        switch (menuheader.USE.trim()) {
            case "T":
                appx_session.currentmenuitems.toolbar[menuType] = appx_session.menucache.currentmenuitems.toolbar[menuType];
                appx_session.currentmenuitems.dropdown[menuType] = appx_session.menucache.currentmenuitems.dropdown[menuType];
                appx_session.currentmenuitems.popup[menuType] = appx_session.menucache.currentmenuitems.popup[menuType];
                break;
            default:
        }
    }
    else {
        if (menu.items.length > 0) {
            var i = 1;
            while (menu.items.length > 0) {
                var mx = menu.items.shift();
                mx.widget = new Widget(null, null, mx.data);
                appx_session.currentmenuitems.all.push(mx.widget);
                if (mx.widget.wUsageToolbar == true) {
                    appx_session.currentmenuitems.toolbar[menuType].push(mx.widget);
                }
                if (mx.widget.wUsagePopup == true) {
                    appx_session.currentmenuitems.popup[menuType].push(mx.widget);
                }
                if (mx.widget.wUsageMenu == true) {
                    appx_session.currentmenuitems.dropdown[menuType].push(mx.widget);
                }
            }
        }
    }
}

//for debug purpose
function buildMenuList(a) {
    for (var i = 0; i < a.length; i++) {
        var b = $("<div style='float:left; border: 3px solid #eee; width: 18%; height: 25px; padding: 5px;'>").html(a[i].wLabel);
        var c = $("<div style='float:left; border: 3px solid #eee; width: 18%; height: 25px; padding: 5px;'>").html(a[i].wGroupName);
        var d = $("<div style='float:left; border: 3px solid #eee; width: 18%; height: 25px; padding: 5px;'>").html(a[i].wUsagePopup);
        var e = $("<div style='float:left; border: 3px solid #eee; width: 18%; height: 25px; padding: 5px;'>").html(a[i].wUsageMenu);
        var f = $("<div style='float:left; border: 3px solid #eee; width: 18%; height: 25px; padding: 5px;'>").html(a[i].wCommand);
        $("#menu").append(b);
        $("#menu").append(c);
        $("#menu").append(d);
        $("#menu").append(e);
        $("#menu").append(f);
        $("#menu").append("<br>\n");
    }
}

function fixMenuLabelsAndGroups(m, defaultGroup, menuTree) {
    // This function will recode the menu list to ensure that 
    // duplicate menu group names are unique and that the group
    // children are attached to the correct parent group.  We 
    // do this by adding a prefix of OptNo + : as a prefix to 
    // the group name.  We determine which group a child belongs
    // to by attaching children to the parent group that has the
    // highest OptNo that is less than our child OptNo.  If there
    // isn't one less than our child OptNo then we attach it to
    // the first menu with the right name and the lowest OptNo
    //
    // For example, this raw menu structure...
    //
    // OptNo Label                Group
    // ----- -------------------- --------------------
    //   10  Inputs               Code Files
    //   20  Countries            Inputs
    //   30  Timezones            Inputs
    //   40  Countries            Reports
    //   50  Timezones            Reports
    //   60  Reports              Code Files
    //   70  Inputs               Master Files
    //   80  Reports              Master Files
    //   90  Customers            Inputs
    //  100  Brokers              Inputs
    //  110  Customers            Reports
    //  120  Brokers              Reports
    //
    // Gets some group name changes to Become this...
    //
    // OptNo Label                Group
    // ----- -------------------- --------------------
    //   10  10:Inputs            Code Files
    //   20  Countries            10:Inputs
    //   30  Timezones            10:Inputs
    //   40  Countries            60:Reports
    //   50  Timezones            60:Reports
    //   60  60:Reports           Code Files
    //   70  70:Inputs            Master Files
    //   80  80:Reports           Master Files
    //   90  Customers            70:Inputs
    //  100  Brokers              70:Inputs
    //  110  Customers            80:Reports
    //  120  Brokers              80:Reports
    //
    // Which gets processed from top to bottom to create a one level deep nested array like this...
    //
    // menu[Files       ] = []
    // menu[Code Files  ] = [ 10:Inputs,      60:Reports    ]
    // menu[10:Inputs   ] = [ Countries(20),  Timezones(30) ]
    // menu[60:Reports  ] = [ Countries(40),  Timezones(50) ]
    // menu[Master Files] = [ 70:Inputs,      80:Reports    ]
    // menu[70:Inputs   ] = [ Customers(90),  Brokers(100)  ]
    // menu[80:Reports  ] = [ Customers(110). Brokers(120)  ]
    // menu[Options     ] = []
    // menu[Help        ] = []
    //
    // Which gets processed recursively by index number to become this...
    //
    // Files
    // Code Files
    //   10:Inputs
    //     Countries(20)
    //     Timezones(30)
    //   60:Reports
    //     Countries(40)
    //     Timezones(50)
    // Master Files
    //   70:Inputs
    //     Customers(90)
    //     Brokers(100)
    //   80:Reports
    //     Customers(110)
    //     Brokers(120)
    // Options
    // Help
    //
    // Which will end up getting rendered like this...
    //
    //  [Files]   [Code Files]   [Master Files]   [Options]   [Help]
    //              10:Inputs      70:Inputs
    //                Countries      Customers
    //                Timezones      Brokers
    //              60:Reports     80:Reports
    //                Countries      Customers
    //                Timezones      Brokers

    //*********************************
    // 1 - Create a list of menu groups
    //*********************************

    var groupCount = 0;
    var workGroups = [];

    // Add known menu groups
    workGroups["file"] = [];
    workGroups["options"] = [];
    workGroups["help"] = [];
    workGroups[defaultGroup] = [];

    defaultGroupLC = defaultGroup.toLowerCase();

    // Add group names from menu item widgets
    for (var i = 0; i < m.length; i++) {
        if (m[i].wGroupName && m[i].wGroupName != "") {
            var name = m[i].wGroupName.toLowerCase();
            m[i].menuGroup = name;
            workGroups[name] = [];
        }
        else {
            m[i].menuGroup = defaultGroupLC;
        }

        if (m[i].wLabel && m[i].wLabel != "") {
            var name = m[i].wLabel.toLowerCase();
            m[i].menuLabel = name;
        }
        else
            m[i].menuLabel = null;
    }

    groupCount = workGroups.length;

    //***********************************************************
    // 2 - Find widgets that define these groups and adjust names
    //***********************************************************

    // If found add Command number to list for dups and adjust names to be unique
    var topLevelMenus = {};

    for (var i = 0; i < m.length; i++) {
        if (m[i].menuLabel) {
            if (workGroups[m[i].menuLabel]) {
                workGroups[m[i].menuLabel].push(parseInt(m[i].wCommand));
                m[i].menuLabel = ":" + m[i].wCommand + ":" + m[i].menuLabel;

                if (m[i].wGroupName == null)
                    topLevelMenus[m[i].menuLabel] = true;
            }
        }
    }

    for (var key in workGroups) {
        var group = workGroups[key];
        if (group.length == 0)
            topLevelMenus[key] = true;
    }

    //**************************************************************
    // 3 - Find widgets that belong to these groups and adjust names
    //**************************************************************
    // Find menu items that link to these groups and adjust their group names
    // For dups, a child belongs to the group that it's command is less that or equal to.
    // It there are none less than or equal then use the first one in the list.
    for (var i = 0; i < m.length; i++) {
        if (m[i].menuGroup) {
            var name = m[i].menuGroup;
            var group = workGroups[name];
            if (group) {
                var found = -1;
                for (var j = 0; j < group.length; j++) {
                    if (group[j] <= parseInt(m[i].wCommand)) {
                        found = j;
                    }
                }

                if (found < 0 && group.length > 0)
                    found = 0;

                if (found >= 0) {
                    m[i].menuGroup = ":" + group[found] + ":" + m[i].menuGroup;
                }
            }
        }
    }

    var menuGroupNames = [];

    // Add known menu groups
    menuGroupNames["file"] = [];
    menuGroupNames["options"] = [];
    menuGroupNames["help"] = [];

    // Add group names from menu item widgets
    for (var i = 0; i < m.length; i++) {
        if (m[i].menuGroup) {
            menuGroupNames[m[i].menuGroup] = [];
        }
    }

    //**************************************************************
    // 4 - Create a flat list of menu groups and their children
    //**************************************************************
    var menuGroups = {};
    var topMenuList = [];

    for (var i = 0; i < m.length; i++) {
        var lname = m[i].menuLabel;
        var gname = m[i].menuGroup;
        if (lname) {
            if (gname) {
                var mgroup = menuGroups[gname];
                if (!mgroup) {
                    menuGroups[gname] = new mGroup(gname, m[i].wGroupName, null);
                    mgroup = menuGroups[gname];
                    if (topLevelMenus[gname])
                        topMenuList.push(gname);
                }
            }
            mgroup.children.push(new mChild(lname, m[i].wLabel, m[i]));
            var mgn = menuGroupNames[lname];
            if (mgn) {
                if (menuGroups[lname]) {
                    menuGroups[lname].widget = m[i];
                }
                else {
                    menuGroups[lname] = new mGroup(lname, m[i].wLabel, m[i]);
                    if (topLevelMenus[lname])
                        topMenuList.push(lname);
                }
            }
        }
    }

    //**************************************************************
    // 5 - Turn the flat list into an indented structure to return
    //**************************************************************

    function menuChild(menuName) {
        var mgroup = menuGroups[menuName];
        var menuNode = new MenuNode(mgroup.label, mgroup.widget);
        for (var nodeNdx = 0; nodeNdx < mgroup.children.length; nodeNdx++) {
            var mchild = mgroup.children[nodeNdx];
            if (menuGroups[mchild.menuLabel])
                menuNode.children.push(menuChild(mchild.menuLabel));
            else
                menuNode.children.push(new MenuItem(mchild.label, mchild.widget));
        }
        return menuNode;
    }

    for (var topNdx = 0; topNdx < topMenuList.length; topNdx++) {
        var menuName = topMenuList[topNdx];
        if (menuGroups[menuName].children.length > 0) {
            var menuNode = menuChild(menuName);
            if (menuName == 'file')
                menuTree.fileMenu = menuNode;
            else if (menuName == 'options')
                menuTree.optnMenu = menuNode;
            else if (menuName == 'help')
                menuTree.helpMenu = menuNode;
            else
                menuTree.userMenu.push(menuNode);
        }
    }
}

function getMenuGroups(m) {
    for (var i = 0; i < m.length; i++) {
        if (m[i].wGroupName) {
            if (m[i].wGroupName != "" && menugroups.indexOf(m[i].wGroupName) < 0) {
                menugroups.push(m[i].wGroupName);
                menu[m[i].wGroupName] = {};
                menu[m[i].wGroupName].groups = [];
                menu[m[i].wGroupName].items = [];
                menu[m[i].wGroupName].level = 1;
                menu[m[i].wGroupName].isGroup = true;
                menu[m[i].wGroupName].parent = m[i].wGroupName;
            }
        }
        else {
            //top level
            menu[m[i].wLabel] = {};
            menu[m[i].wLabel].groups = [];
            menu[m[i].wLabel].level = 0;
        }
    }

    //process menu headerdata
    var mheaderitem = appx_session.currentmenu.header;
    for (var mh = 0; mh < mheaderitem.length; mh++) {
        if (menugroups.indexOf(mheaderitem[mh].wLabel) < 0) {
            menugroups.push(mheaderitem[mh].wLabel);
            menu[mheaderitem[mh].wLabel] = {};
            menu[mheaderitem[mh].wLabel].groups = [];
            menu[mheaderitem[mh].wLabel].items = [];
            menu[mheaderitem[mh].wLabel].level = 1;
            menu[mheaderitem[mh].wLabel].isGroup = true;
            menu[mheaderitem[mh].wLabel].parent = mheaderitem[mh].wLabel;
        }
    }

    menu.groups = menugroups;
}

function getMenuLabels(m) {
    for (var i = 0; i < m.length; i++) {
        if (m[i].wLabel) {
            if (m[i].wLabel != "") {
                menulabels.push(m[i].wLabel);
                if (menugroups.indexOf(m[i].wLabel) > -1) {
                    menu[m[i].wLabel].isGroup = true;
                    if (m[i].wGroupName && m[i].wGroupName != "") {
                        menu[m[i].wLabel].parent = m[i].wGroupName;
                    }
                    else {
                        menu[m[i].wLabel].parent = m[i].wLabel;
                    }
                    menu[m[i].wLabel].command = m[i].wCommand;
                }
                if (m[i].wGroupName && m[i].wGroupName != "" && menugroups.indexOf(m[i].wGroupName) > -1) {
                    menu[m[i].wGroupName].items.push(m[i]);
                }
                if ((m[i].wGroupName == null || m[i].wGroupName == "") && menugroups.indexOf(m[i].wLabel) < 0) {
                    menu[new Widget(null, null, appx_session.currentmenu.block.data.headerdata).wLabel].items.push(m[i]);
                }
            }
        }
    }
}

function setGroupHier(g, m) {
    for (var i = 0; i < m.length; i++) {
        if (m[i].wGroupName) {
            if (m[i].wGroupName != "" && menugroups.indexOf(m[i].wGroupName) < 0) {
                menugroups.push(m[i].wGroupName);
            }
        }
    }
}

function buildMenuHTMLtree(menuNode) {
    if (menuNode == null || menuNode.children.length == 0)
        return "";
    var cmd = (menuNode.widget == null ? '' : " onclick='appxwidgetcallback(" + menuNode.widget.wCommand + ")'");
    var wHTML = "<li" + cmd + ">" + (menuNode.widget == null ? menuNode.label : menuNode.widget.wLabel) + "<ul>";
    for (var ndx = 0; ndx < menuNode.children.length; ndx++) {
        var menuObj = menuNode.children[ndx];
        if (menuObj instanceof MenuNode) {
            wHTML += buildMenuHTMLtree(menuObj);
        }
        else {
            var wgt = (menuObj.widget == null ? false : true);
            var cmd = null;
            var lbl = null;
            //TODO: Fix ME add a class for enabled menu items(.appx-menu-item), not an inline style
            var style = " style='background-position:2% center;background-repeat:no-repeat;background-size:14px 14px;padding-left: 25px;'";
            var sepbefore = false;
            var sepafter = false;
            var disabled = '';
	    var tbvisible = true;

            if (wgt) {
		var tbvisible = menuObj.widget.wVisible == null ? true : menuObj.widget.wVisible;
                cmd = " onclick='appxwidgetcallback(" + menuObj.widget.wCommand + ")'";
                if (cmd && menuObj.widget.wCommand == "20032") {
                    cmd = " onclick='appx_session.showPreferences()' ";
                }
                lbl = menuObj.widget.wLabel;
                sepbefore = menuObj.widget.wSepBefore == null ? false : true;
                sepafter = menuObj.widget.wSepAfter == null ? false : true;
                disabled = menuObj.widget.wEnabled == false ? ' disabled' : '';
                //TODO: Fix ME add a class for disabled menu items(.appx-menu-item .appx-menu-item-disabled), not an inline style
                style = menuObj.widget.wEnabled == false ? " style = 'background-position:2% center;background-repeat:no-repeat;background-size:14px 14px;padding-left: 25px;color: #eee;text-shadow: 1px 1px #333'" : style;
                cmd = menuObj.widget.wEnabled == false ? "" : cmd;
            }
            else {
                lbl = menuObj.label;
            }

	    if (tbvisible) {
		var li = $("<li" + cmd + "" + style + "" + disabled + ">");
		$(li).text(lbl);
		if (sepbefore) $(li).addClass("sepbefore");
		if (sepafter) $(li).addClass("sepafter");
		if (disabled) $(li).prop("disabled", true);
		var icon = (wgt == false ? false : menuObj.widget.wIconEnabled == null ? false : $(li).addClass(AppxResource.load(menuObj.widget.wIconEnabled)));
		wHTML += $(li)[0].outerHTML;
	    }
        }
    }
    wHTML += "</ul></li>";
    return wHTML;
}

function buildMenuHTML(menuTree) {
    $("#appx-main-nav").html("");
    var menunav = $("#appx-main-nav");
    var topMenu = $("<ul>").attr("id", "topMenu").css({
        "width": $("#appx_main_container").width()
    }).appendTo(menunav);
    var menuHTML = "";
    menuHTML += buildMenuHTMLtree(menuTree.fileMenu);
    menuHTML += buildMenuHTMLtree(menuTree.optnMenu);
    for (var ndx = 0; ndx < menuTree.userMenu.length; ndx++)
        menuHTML += buildMenuHTMLtree(menuTree.userMenu[ndx]);
    menuHTML += buildMenuHTMLtree(menuTree.helpMenu);
    $(topMenu).append(menuHTML);
}

function buildPopupMenuHTML(menuTree) {
    $("#appx_popup_nav").html("");
    var menunav = $("#appx_popup_nav");
    var popupMenu = $("<ul>").attr("id", "popupMenu").appendTo(menunav);
    var menuHTML = "";
    menuHTML += buildMenuHTMLtree(menuTree.fileMenu);
    menuHTML += buildMenuHTMLtree(menuTree.optnMenu);
    for (var ndx = 0; ndx < menuTree.userMenu.length; ndx++)
        menuHTML += buildMenuHTMLtree(menuTree.userMenu[ndx]);
    menuHTML += buildMenuHTMLtree(menuTree.helpMenu);
    $(popupMenu).append("<li><ul><li>" + menuHTML + "</li></ul></li>");
}

function createDropDownMenu() {
    menugroups = [];
    menulabels = [];
    menuitems = [];
    menu = {};
    menu2 = {};
    var menuhtml = "";
    var menus = [null, null, null];
    var names = ["Proc", "Appl", "Sys"];
    var menuTree = new MenuTree();
    for (var menuType = 2; menuType >= 0; menuType--) {
        var a = appx_session.currentmenuitems.dropdown[menuType];
        if (a.length > 0) {
            menus[menuType] = fixMenuLabelsAndGroups(a, names[menuType], menuTree);
        }
    }
    buildMenuHTML(menuTree);
}

function createPopupMenu() {
    menugroups = [];
    menulabels = [];
    menuitems = [];
    menu = {};
    menu2 = {};
    var menuhtml = "";
    var menus = [null, null, null];
    var names = ["Proc", "Appl", "Sys"];
    var menuTree = new MenuTree();
    for (var menuType = 2; menuType >= 0; menuType--) {
        var a = appx_session.currentmenuitems.popup[menuType];
        if (a.length > 0) {
            menus[menuType] = fixMenuLabelsAndGroups(a, names[menuType], menuTree);
        }
    }
    buildPopupMenuHTML(menuTree);
}

function createdToolbarMenu() {
    var tbi = [];
    //no need to fix toolbar item hierarchy
    for (var menuType = 2; menuType >= 0; menuType--) {
        var a = appx_session.currentmenuitems.toolbar[menuType];
        for (var i = 0; i < appx_session.currentmenuitems.toolbar[menuType].length; i++) {
            tbi.push(appx_session.currentmenuitems.toolbar[menuType][i]);
        }
    }

    var $newtoolbar = $("<div>");
    var $newdeftoolbar = $("<div>");
    while (tbi.length > 0) {
        var mx = {};
        mx.widget = tbi.shift();
        if (mx.widget.wIconEnabled) {
            var item = $('<div style="cursor:pointer; text-align:center; vertical-align: bottom; float:left;height:40px;width:50px;color:black;border: 0px solid #333;background-position:50% 20%;; background-size:50%;">');
            if (mx.widget.wShortLabel) {
                $(item).html('<span style="vertical-align: -27px;font-size:11px;font-family:arial">' + mx.widget.wShortLabel.trim() + '</span>');
            }
            var cacheid = AppxResource.load(mx.widget.wIconEnabled);
            $(item).addClass("appxitem"); //item
            $(item).addClass("toolbaritem");
            $(item).addClass(cacheid);

            var tbenabled = mx.widget.wEnabled == null ? true : mx.widget.wEnabled;

            if (tbenabled == false) {
                $(item).addClass("tbdisabled");
            }

            $(item).attr("id", addClientId(mx.widget.wCommand,mx.widget.wClientId));


            $(item).click(function () {
                appxwidgetcallback(this.id);
            });


            if (mx.widget.wTooltip) {
                $(item).prop('title', mx.widget.wTooltip);
                $(item).tooltip({
                    content: function () {
                        return this.title();
                    }
                });
            }
            else if (mx.widget.wLabel) {
                $(item).prop('title', mx.widget.wLabel);
                $(item).tooltip({
                    content: function () {
                        return this.title();
                    }
                });
            }

            var tbvisible = mx.widget.wVisible == null ? true : mx.widget.wVisible;

            if (tbvisible) {
                if (mx.widget.wCommand > "250") {
                    //FIX ME: deep clone keeps event, don't have to add em back with .clone(true)
                    //TODO: Fix clone
                    var ic = $(item).clone();
                    $(ic).attr("id", addClientId("sk_" + mx.widget.wCommand,mx.widget.wClientId));
                    $(ic).css({
                        "background-color": "transparent"
                    });
                    $(ic).click(function () {
			    var i = getClientId(this.id).replace("sk_", "");
                        appxwidgetcallback(i);
                    });
                    $newdeftoolbar.append(ic);
                }
                else {
                    $newtoolbar.append(item);
                }
            }
        }
    }
    $("#default-toolbar").html($newdeftoolbar);
    $("#toolbar").html($newtoolbar);
}