/*********************************************************************
 **
 **   server/appx-client-screen.js - Client Screen processing
 **
 **   This module contains code to process client resources.
 **
 *********************************************************************/

// what_str =  "@(#)Appx $Header: /src/cvs/appxHtml5/server/appx-client-screen.js,v 1.177 2015/06/26 20:18:57 pete Exp $";

var screenflipped = false;
var blocked = false;
var inProgressBox = null;

function appxmsgshandler(x) {
    appx_session.msgs = [];
    var i = 0;
    var $msgshtml = "";
    while (x.data.length > 0) {
        var msg = x.data.shift();
        if (msg.txtlen > 0 && msg.txtval.indexOf("Building Scan List") < 0) {
            var $msghtml = $("<span>").html(msg.txtval);
            $msghtml.addClass("status-msg");
            switch (msg.severity) {
                case 0:
                    $msghtml.addClass("status-msg-info");
                    break;
                case 1:
                    $msghtml.addClass("status-msg-warning");
                    break;
                case 2:
                    $msghtml.addClass("status-msg-error");
                    break;
                case 3:
                    $msghtml.addClass("status-msg-cancel");
                    break;
            }
            appx_session.msgs.push($msghtml);
        }
    }
}

function applymessages() {
    var $newhtml = $("<div>");
    var msgcount = 0;
    for (var i = 0; i < appx_session.msgs.length; i++) {
        var $msghtml = appx_session.msgs[i];
        if (i > 0)
            $newhtml.append("<br>");
        $newhtml.append($msghtml);
    }
    $("#appx-status-msg").html($newhtml);
}

function appxfindbox(pos_row, pos_col, size_rows, size_cols, includeScroll) {
    var ret = null;
    var boxes = appx_session.current_show.boxes;
    for (var boxIdx = 0; boxIdx < boxes.length; boxIdx++) {
        var box = boxes[boxIdx];
        if (pos_col >= box.begin_column && pos_col <= box.end_column && pos_row >= box.begin_row && pos_row <= box.end_row) {
            if (pos_row + size_rows - 1 <= box.end_row && pos_col + size_cols - 1 <= box.end_column) {
                if (includeScroll == true || ( appxIsScroll(box) == false && appxIsScrollReg(box) == false ))
                    ret = box;
            }
        }
    }
    return ret;
}

function applyimage(key) {
    if (key != null) {
        var url = appx_session.image_cache[key].url;
        if (url != "") {
            //feed img tags instead of background-image
            $("." + key + "_img").attr("src", url);
            //feed fake attributes for mouse hovers
            $("." + key + "_imgRO").attr("srcRO", url);
            // feed background image css
            $("." + key).css({
                "background-image": "url('" + url + "')"
            });
        }
    }
    if (appx_session.pendingResources.length == 0) {
        appxSetStatusStateText(APPX_STATE_READY);
        if (!screenflipped)
            appxshowscreen();
    }
    else {
        appxSetStatusStateText(APPX_STATE_IMAGES);
    }
}

function applystyles() {
    try {
        appx_session.applyStylesCount++;
        if (appx_session.image_cache) {
            for (var i = 0; i < appx_session.image_cache.keys.length; i++) {
                var key = appx_session.image_cache.keys[i];
                applyimage(key);
            }
        }
        if (appx_session.applyStylesCount > 1)
            return;
        appxApplyStylesCheckbox();
        appxApplyStylesColorPicker();
        appxApplyStylesDate();
        appxApplyStylesEditor();
        appxApplyStylesHtmlViewer();
        appxApplyStylesTable();
	appxApplyStylesTitleButtons();
        $('#appx_main_container').contextMenu('#appx_popup_nav');
    }
    catch (ex) {
        console.log("applystyles: " + ex);
    }
}

function appxApplyStylesTitleButtons() {
    $.each( $(".appx-title-button-help"), function (k, el0) {
		$(el0).click(function (event) {
			event.preventDefault();
			appxwidgetcallback(OPT_WHATS_THIS);
		    });
	    });
    $.each( $(".appx-title-button-ok"), function (k, el0) {
		$(el0).click(function (event) {
			event.preventDefault();
			appxwidgetcallback(OPT_ENTER);
		    });
	    });
    $.each( $(".appx-title-button-close"), function (k, el0) {
		$(el0).click(function (event) {
			event.preventDefault();
			appxwidgetcallback(OPT_END);
		    });
	    });
}

function appxApplyStylesCheckbox() {
    try {
        $.each($("input[type=checkbox]"), function (k, el0) {
		$(el0).data('checked', $(el0).checked)
		    .keypress(function(event) {
			    el = $(this);
			    if( event.keyCode == 110 || event.keyCode == 78 ) { // n or N
				el.data('checked', 0);
				el.prop('indeterminate', false);
				el.prop('checked', false);
				el.val("N");
			    }
			    else if( event.keyCode == 121 || event.keyCode == 89 ) { // y or Y
				el.data('checked', 2);
				el.prop('indeterminate', false);
				el.prop('checked', true);
				el.val("Y");
			    }
			})   
		    .click(function (e) {
			    el = $(this);
			    if (el.data('checked') == 0 && el.hasClass('appx-nullok') == false)
				el.data('checked', 1);
			    switch (el.data('checked')) {
				// unchecked, going indeterminate  
			    case 0:
				el.data('checked', 1);
				el.prop('indeterminate', true);
				el.val(" ");
				break;
				// indeterminate, going checked  
			    case 1:
				el.data('checked', 2);
				el.prop('indeterminate', false);
				el.prop('checked', true);
				el.val("Y");
				break;
				// checked, going unchecked  
			    default:
				el.data('checked', 0);
				el.prop('indeterminate', false);
				el.prop('checked', false);
				el.val("N");
			    }
			});
	    });
    }
    catch (ex) {
        console.log("appxApplyStylesCheckbox: " + ex);
    }
}

function appxApplyStylesColorPicker() {
    try {
        //may need to loop and get ids here;
        $.each($(".appxcolorpicker"), function (i, el) {
            $(el).parent().width($(el).parent().width() + (appx_session.colwidthpx * 2) - 2);
            $(el).colpick({
                layout: 'hex',
                submit: true,
                colorScheme: 'light',
                onChange: function (hsb, hex, rgb, elc, bySetColor) {
                    $(elc).css('border-color', '#' + hex);
                    // Fill the text box just if the color was set using the picker, and not the colpickSetColor function.
                    if (!bySetColor) $(elc).val(hex);
                }
            }).keyup(function () {
                $(this).colpickSetColor(this.value);
            });
            //icon click opens colorpicker
            $(el).parent().find("img").colpick({
                layout: 'hex',
                submit: true,
                colorScheme: 'light',
                onChange: function (hsb, hex, rgb, elc, bySetColor) {
                    $(elc).parent().find("input").css('border-color', '#' + hex);
                    // Fill the text box just if the color was set using the picker, and not the colpickSetColor function.
                    if (!bySetColor) $(elc).parent().find("input").val(hex);
                }
            });
        });
    }
    catch (ex) {
        console.log('appxApplyStylesColorPicker: ' + ex);
    }
}

function appxApplyStylesDate() {
    try {
        $.each($(".appxdatefield"), function (i, el) {
            $(el).width($(el).width() + (appx_session.colwidthpx * 2));
            var datevalue = $(el).find(".appxdatevalue").first();
            datevalue.css({
                "font-size": appx_session.basefontsize + "px"
            });
            var datepicker = $(el).find(".appxdatepicker").first();
            var parts = JSON.parse(datepicker.val());
            var opt = {
                beforeShow: function (inp, inst) {
                    appx_session.activeDatepicker = $(this);
                },
                buttonImage: ""+appxClientRoot+"/images/calendar.gif",
                buttonImageOnly: true,
                changeMonth: true,
                changeYear: true,
                dateFormat: parts.datemsk,
                onClose: function (date, obj) {
                    //appx_session.activeDatepicker = null;
                    var ad = APPX.DateFormatter2(
                        date, obj.settings.dateFormat, obj.settings.timeFormat
                    );
                    var ar = $(this).closest(".appxitem").attr('id').split('_');
                    sendappxdate(parseInt(ar[2]), parseInt(ar[1]), ad);
                },
                showButtonPanel: true,
                showOn: "button",
                timeFormat: parts.timemsk,
                yearRange: "1000:3000"
            };
            datepicker.datepicker(opt);
            datepicker.val(parts.value);
            $(el).find("img").css({
                "position": "absolute",
                "right": "2px",
                "top": "2px",
                "z-index": 100000
		    }).click(function(){
			    appx_session.activeDatepicker = $(this);			    
			});

	    datevalue.on("input keydown", function () {
		    $(this).addClass("dirty");
		    appxSetStatusStateText(APPX_STATE_DIRTY);
		});

        });
    }
    catch (ex) {
        console.log('appxApplyStylesDate: ' + ex);
    }
}

function appxApplyStyleEditor(el) {
    try {
            var twidth = $(el).css("width");
            var theight = $(el).css("height");
            // the height of the CKE editor greatly depends on the editor's config settings
            // We should plan on a better way to figure exact height rather than subtract the header height
	    var config = {
                toolbar: [
                    ['Font', 'FontSize'],
                    ['TextColor', 'BGColor', 'Bold', 'Italic', '-', 'Source'], 
		    '/', 
		    ['JustifyLeft', 'JustifyCenter', 'JustifyRight', 'JustifyBlock', '-', 'NumberedList', 'BulletedList', '-', 'Outdent', 'Indent'],
                    ['Cut', 'Copy', 'Paste', '-', 'Undo', 'Redo', '-', 'Link', 'Table', 'HorizontalRule']
                ]
            };

	    if( $(el).data("editorConfig")) {
		config = AppxResource.cache[$(el).data("editorConfig")];		
	    }

	    if ($(el).attr("disabled") == "disabled") {
		config.removePlugins = "elementspath,toolbar";
		config.resize_enabled = false;
	    }

	config.bodyClass = "appx-editor-body";
	//EDITOR.addCss("body { margin: 2px; }");
        var editor = CKEDITOR.replace($(el).attr("id"), config );
	
	    $(el).attr("name",editor.name);
            editor.on('instanceReady', function () {
		$("#cke_"+$(el).attr("id")).find("iframe").attr("title","");
		$("#cke_"+$(el).attr("id")).css({
			"position": "absolute",
			"top": $(el).css("top"),
			"left": $(el).css("left"),
			"z-index": $(el).css("z-index")
			    });
		setTimeout(function() {
                    editor.resize(
			      ( parseInt(twidth) -4 
				), 
			      ( parseInt(theight) +2
			      ), false);
		    },0);
		});
	    editor.resetDirty();
    }
    catch (ex) {
        console.log('appxApplyStylesEditor: ' + ex);
    }
}

function appxApplyStylesEditor() {
    try {
        //may need to loop and get ids here;
        $.each($(".ckeditor"), function (i, el) {
		if( $(el).data("editorConfig") && !AppxResource.cache[$(el).data("editorConfig")] ) {
		    // do nothing yet, it will load after the resource arrives
		}
		else {
		    appxApplyStyleEditor(el);
		}
	});
    }
    catch (ex) {
        console.log('appxApplyStylesEditor: ' + ex);
    }
}

function appxApplyStylesHtmlViewer() {
    try {
        $.each($(".appx-html-viewer a"), function (k, el0) {
            var opt = $(el0).attr("href");
            if (!isNaN(opt)) {
                $(el0).attr({
                    "href": "#"
                }).click(function (event) {
                    event.preventDefault();
                    appxwidgetcallback(parseInt(opt));
                    return false;
                });
            }
        });
    }
    catch (ex) {
        console.log('appxApplyStylesHtmlViewer: ' + ex);
    }
}

function appxApplyStylesTable() {
    try {
        $.each($(".appxtablewidget"), function (i, el) {
            var tabledata = appx_session.currenttabledata[$(el).attr("id")].tabledata;
            createGrid(appx_session.currenttabledata[$(el).attr("id")].tabledata[0], tabledata, $(el));
            var gridprops = null;
            if (appx_session.hasOwnProperty("gridpropscache"))
                gridprops = appx_session.gridpropscache[$(el).data("tableHashKey")];
            if (gridprops) {
                var uigrid = $("#newtablewidget_" + $(el).attr('id'));
		if( gridprops.sortcol ) {
		    uigrid.setGridParam({sortorder:gridprops.sortorder}).jqGrid("sortGrid",gridprops.sortcol,true);
		}
                $(el).find(".ui-jqgrid-bdiv").scrollTop(gridprops.scrollTop);
                $(el).find(".ui-jqgrid-bdiv").scrollLeft(gridprops.scrollLeft);
                uigrid.resetSelection();
                if (gridprops.selrows && gridprops.selrows.length > 0) {
                    for (var i = 0; i < gridprops.selrows.length; i++)
                        uigrid.setSelection(gridprops.selrows[i], true);
                }
            }
	    $(el).height((parseInt($(el).height())+appx_session.rowheightpx)+"px");
        });
    }
    catch (ex) {
        console.log('appxApplyStylesTable: ' + ex);
    }
}


//Attributes Message Handler
function appxattributeshandler(x) {}

//Extra Attributes Message Handler
function appxextraattributeshandler(x) {}

//Screen Message Handler
function appxscreenhandler(x) {
    var screendatastr = "";
    //TODO? not adding screenBuf to the DOM might be faster because of reflow (replaceWith)
    if ($_screenBuf == null) {
        var $screen = $("#appx_main_container").empty();
        $_screenBuf = $screen.clone();
        $_screenBuf.attr("id", "screenBuf");
        $_screenBuf.appendTo($screen.parent());
        $_screenBuf.hide();
    }
    $_screenBuf.html("<div id='appx_main_background' style='float:left;width:70%;'>" + screendatastr + "</div>");
    //the menu may have multiple blocks and has to be cleared as a function of the current show/screen
    initializeMenu();
}

//raw screen data not used at the moment
function appxscreenhandlerdata(x) {
    var screendata = [];
    var screendatastr = "";
    var datalength = 0;
    while (x.length > 0) { //process compressed screen
        var mask = x.shift();
        logactivity("mask:  " + mask);
        if (mask > 255) {
            screendata = new Array(mask);
            var rptchar = x.shift();
            for (var i = 0; i < screendata.length; i++) {
                screendata[i] = rptchar;
            }
            x = x.slice(screendata.length, x.length);
        }
        else {
            x = x.slice(1, x.length);
            datalength = x.shift() + 1;
            screendata = x.slice(0, datalength);
            screendatastr += ab4str(screendata);
            x = x.slice(datalength, x.length);
        }
    }
    return screendatastr;
}

function appxMergeBoxes(oldbox, newbox) {
    var topAdj = oldbox.begin_row - newbox.begin_row;
    var leftAdj = oldbox.begin_column - newbox.begin_column;
    var topAdjPx = topAdj * appx_session.rowheightpx;
    var leftAdjPx = leftAdj * appx_session.rowwidthpx;
    while (oldbox.items.length) {
        var boxitem = oldbox.items.shift();
        var item = boxitem[0];
        var $tag = boxitem[1];
        item.pos_row += topAdj;
        item.pos_col += leftAdj;
        $tag.css({
            "top": (parseInt($tag.css("top")) + topAdjPx) + "px"
        });
        $tag.css({
            "left": (parseInt($tag.css("left")) + leftAdjPx) + "px"
        });
        newbox.items.push(boxitem);
    }
    while (oldbox.widgets.length) {
        var boxwdgt = oldbox.widgets.shift();
        var wx = boxwdgt[0];
        var $tag = boxwdgt[1];
        wx.wPositionY += topAdj;
        wx.wPositionX += leftAdj;
        $tag.css({
            "top": (parseInt($tag.css("top")) + topAdjPx) + "px"
        });
        $tag.css({
            "left": (parseInt($tag.css("left")) + leftAdjPx) + "px"
        });
        newbox.widgets.push(boxwdgt);
    }
    while (oldbox.rowtext.length) {
        var rowtext = oldbox.rowtext.shift();
        rowtext.pos_row += topAdj;
        rowtext.pos_col += leftAdj;
        newbox.rowtext.push(rowtext);
    }
}

var _boxcur = null; //appx_session?
function appxshowboxes() {
    try {
        var box1 = null,
            box2 = null,
            box3 = null;
        var boxcnt = 0;
        var $boxcurhtml = null;
        var cel = appxGetCellSize();
        var cur = appxGetCursorPos();
        var bPrevScrollAct = false;
        var topBoxIdx = 0;
        appx_session.topbox = null;
        appx_session.tablist = null;
        for (var i = 0; i < appx_session.current_show.boxes.length; i++) {
            inProgressBox = null;
            var box = appx_session.current_show.boxes[i];
            //console.log("box %d(%d)(%d) begin: %d/%d end %d/%d mask 0x%s, label=%s",i,(box.widget == null ? -1 : box.widget.wBoxNumber),box.newbox,box.begin_row,box.begin_column,box.end_row,box.end_column,box.bit_mask.toString(16),box.widget.wLabel);

            if (appxIsInProgress(box))
                inProgressBox = box;
            if (appxIsEmBuild(box)) {
                $("#appx_status_date").html(box.widget.wLabel);
                continue;
            }
            if (box.begin_row == 0 && box.end_row == 0) {
                continue;
            }
            if (box.end_row > (appx_session.screenrows - 3))
                box.end_row = appx_session.screenrows - 3;
            if (box.end_column > appx_session.screencols)
                box.end_column = appx_session.screencols;
            // Full window scrolling inputs, not really scrolling but used to get Prev/Next Record with PageUp/Down keys
            if (box1 != null && appxIsScrollReg(box) && appxIsScrollAct(box) && appxIsScroll(box1) && box.begin_row == box1.begin_row && box.end_row == box1.end_row && box.begin_column == box1.begin_column && box.end_column == box1.end_column) {
                box.newbox = i - 1;
                box.bit_mask &= ~SCROLL_REG;
                box.bit_mask &= ~SCROLL_ACT;
                box1.bit_mask &= ~SCROLL;
                appxMergeBoxes(box, box1);
                box3 = null;
                box2 = null;
                box1 = null;
            }
            if (appxIsScrollReg(box)) {
                box3 = null;
                box2 = null;
                box1 = null;
            }
            else {
                box3 = box2;
                box2 = box1;
                box1 = box;
            }
            if (box3 != null) {
                if (box3.begin_column == box2.begin_column && box2.begin_column == box1.begin_column) {
                    if (box3.end_column == box2.end_column && box2.end_column == box1.end_column) {
                        if (box1.end_row >= (box2.begin_row - 1) && box3.end_row >= (box1.begin_row - 1)) {
                            if (appxIsScroll(box1) && box3.bit_mask == box2.bit_mask) {
                                box3.end_row = box2.end_row;
                                box2.newbox = i - 2;
                                appxMergeBoxes(box2, box3);
                                box3.scrollrow = box1.begin_row - (box3.begin_row - 1);
                            }
                        }
                    }
                }
            }
            if (box2 != null) {
                if (box2.begin_column <= box1.begin_column) {
                    if (box2.end_column >= box1.end_column) {
                        if (box2.begin_row < (box1.begin_row) && box2.end_row >= (box1.end_row)) {
                            if (appxIsScroll(box1)) {
                                box2.scrollrow = box1.begin_row - (box2.begin_row - 1);
                            }
                        }
                    }
                }
            }
            if (!(appxIsScroll(box) || appxIsScrollReg(box))) {
                topBoxIdx = i;
            }
        }
        if (appx_session.current_show.boxes[topBoxIdx] && appx_session.current_show.boxes[topBoxIdx].hasOwnProperty("newbox"))
            topBoxIdx = appx_session.current_show.boxes[topBoxIdx].newbox;
        appx_session.topboxid = topBoxIdx;
        var scrollBoxNo = 0;
        var boxLayer = 0;
        for (var i = 0; i < appx_session.current_show.boxes.length; i++) {
            var box = appx_session.current_show.boxes[i];
            if (appxIsEmBuild(box))
                continue;
            if (box.begin_row == 0 && box.end_row == 0) {
                continue;
            }
            if (appxIsInProgress(box))
                continue;
            if (box.newbox >= 0) {
                continue;
            }
            if (box.ignore == true)
                continue;
            if (appxIsScroll(box))
                scrollBoxNo = boxcnt;
            if (appxIsScrollReg(box))
                boxLayer = 1000 * (scrollBoxNo + 1);
            else
                boxLayer = 1000 * (boxcnt + 1);
            box.layer = boxLayer;
            if (!(appxIsScroll(box) || appxIsScrollReg(box))) {
                if (box.widget.wLabel == null && box.rowtext && box.rowtext.length > 0) {
                    var title = "";
                    for (var j = 0; j < box.rowtext.length; j++) {
                        var rowtxt = box.rowtext[j];
                        if (rowtxt.pos_row == box.begin_row) {
                            title += rowtxt.string.trim() + " ";
                            rowtxt.string = "";
                        }
                    }
                    if (title.length > 0)
                        box.widget.wLabel = title;
                }
            }
            var boxwidth = ((box.end_column - box.begin_column) * cel.w) + (3 * cel.w) + 2;
            var boxheight = ((box.end_row - box.begin_row) * cel.h) + (3 * cel.h);
            var boxtop = (box.begin_row * cel.h) - cel.h;
            var boxleft = (box.begin_column * cel.w) - cel.w;
            if (appxIsScroll(box)) {
                boxleft += 3;
                boxwidth -= 6;
            }
            if (appxIsScrollReg(box)) {
                boxleft += 3;
                boxwidth -= 6;
            }
            var $boxhtml2 = null;
            if (box.widget.wLabel && box.widget.wLabel != "") {
                var m = 2; //margin
                var t = m;
                var l = m;
                var w = (boxwidth - (m * 4)) + "px";
                var h = (cel.h) + "px";
                if (appxFillWindow == "true" && box.end_column - box.begin_column + 1 == appx_session.screencols) {
                    w = "100%";
                    t = "0px";
                    l = "0px";
                }
                var $boxhtml2 = $("<div>").css({
                    "border": "0px", //remove any border
                    "box-sizing": "border-box",
                    "display": "table",
                    "height": h,
                    "left": l,
                    "overflow": "hidden",
                    "position": "absolute",
                    "top": t,
                    "width": w,
                    "z-index": boxLayer + 1
                });
                $boxhtml2.addClass("appx-title");
                $boxhtml2.attr("id", "box_2_" + box.widget.wBoxNumber.toString());
                box.widget.isBox = true;
                appxwidgetshandlerpropsex(box.widget, $boxhtml2, true);
            }
            //</toolbar>
            if (appxFillWindow == "true" && box.begin_row == 1 && box.begin_column == 1 && box.end_row == (appx_session.screenrows - 3) && box.end_column == appx_session.screencols) {
                boxwidth = "100%";
            }

            var $boxhtml = $("<div>").css({
                "box-sizing": "border-box",
                "height": boxheight,
                "left": boxleft,
                "overflow": "hidden",
                "position": "absolute",
                "top": boxtop,
                "width": boxwidth,
                "z-index": boxLayer
            });
            if ($boxhtml2) $boxhtml.append($boxhtml2);
            $boxhtml.addClass("appxbox");
            $boxhtml.attr("id", "box_" + box.widget.wBoxNumber.toString());
            if (box.begin_row == 1 && box.begin_column == 1 && box.end_row == (appx_session.screenrows - 3) && box.end_column == appx_session.screencols) {
                $boxhtml.addClass("appxbox-full");
            }
            var titleSave = box.widget.wLabel;
            box.widget.wLabel = null;
            appxwidgetshandlerprops(box.widget, $boxhtml);
            box.widget.wLabel = titleSave;

            //store box between items for tabindexing
            $_screenBuf.append($boxhtml); //"#appx_main_container"
            if ( //lets see if this is the box that contains the cursor
                cur.row >= box.begin_row && cur.row <= box.end_row &&
                cur.col >= box.begin_column && cur.col <= box.end_column) {
                if (_boxcur == null ||
                    //TODO this doesn't work when there are buttons on a footer frame box
                    //-> look at box attributes, ignore scrolling regions?
                    (_boxcur.end_row - _boxcur.begin_row) >= (box.end_row - box.begin_row) ||
                    (_boxcur.end_column - _boxcur.begin_column) >= (box.end_column - box.begin_column)
                ) {
                    if ((box.bit_mask & SCROLL) == 0) {
                        _boxcur = box;
                        $boxcurhtml = $boxhtml;
                    }
                }
            }

            //<scroll>
            //no scrolling classes if box size is full screen
            if (!((box.end_column - box.begin_column) + 1 >= appx_session.screencols &&
                    (box.end_row - box.begin_row) + 1 >= (appx_session.screenrows - 3))) {
                if (appxIsScroll(box)) {
                    $boxhtml.addClass("appx-scroll");
                    $boxhtml.css("top", boxtop + cel.h); //remove the -1 offset row
                    $boxhtml.height(boxheight - (2 * cel.h));
                }
                else if (appxIsScrollReg(box)) {
                    if (appxIsScrollAct(box)) {
                        $boxhtml.addClass("appx-scroll-act");
                        bPrevScrollAct = true;
                    }
                    else {
                        if (bPrevScrollAct) {
                            bPrevScrollAct = false;
                            $boxhtml.addClass("appx-scroll-act-next");
                        }
                        else {
                            $boxhtml.addClass("appx-scroll-reg");
                        }
                        $boxhtml.data("col", box.begin_column);
                        $boxhtml.data("row", box.begin_row);
                    }
                    $boxhtml.click(appxScrollClick);
                    $boxhtml.css("top", boxtop + cel.h); //remove the -1 offset row
                    $boxhtml.height(boxheight - (2 * cel.h)); //remove the default two extra lines
                }
                else if (!box.widget.wBorder) {
                    $boxhtml.addClass("appx-border-bevel-raised");
                }
            }
            //</scroll>

            $boxhtml.dblclick(function () {
		    if (appx_session.getProp("appxDoubleClick") == "true")
			appxwidgetcallback(OPT_ENTER);
            });
            $("a").each(function () {
                $(this).attr("tabindex", -1);
            });
            $("button").each(function () {
                $(this).attr("tabindex", -1);
            });
            appxshowitemsforbox(box, topBoxIdx == i);
            appxshowwidgetsforbox(box, topBoxIdx == i);
            appxshowrowtextforbox(box);
            if (!(appxIsScroll(box) || appxIsScrollReg(box))) {
                if (topBoxIdx != i) {
                    $boxhtml.addClass("appx-not-modifiable");
                }
            }
            boxcnt++;
        }

        if (inProgressBox && !blocked) {
            try {
                blocked = true;
                $("#main").block({
                    message: inProgressBox.widget.wLabel + "...",
                    overlayCSS: {
                        backgroundColor: null,
                        opacity: null,
                        cursor: "wait"
                    },
                    centerY: false,
                    css: {
                        "z-index": 10000
                    },
                    blockMsgClass: "appx-in-progress-msg"
                });
            }
            catch (e) {}
        }
    }
    catch (ex) {
        console.log("appxshowboxes: " + ex);
    }
}

//Display.drawCursor
function appxshowcursor(bFocus) {
    try {
        var bCurItem = false; //cursor inside field?
        var cel = appxGetCellSize();
        var pos = appxGetCursorPos();
        var box = appxfindbox(pos.row, pos.col, 1, 1, true);
        if (!box) return;
        var $cur = $("#appxCursor");
        var firstModItem = null;
        if ($cur.length == 0) {
            $cur = $("<div>");
            $cur.attr("id", "appxCursor");
            $cur.css({
                "background-color": "rgba(0,0,0,.5)",
                "height": cel.h,
                "position": "absolute",
                "width": cel.w,
                "z-index": "100000"
            });
            $("#appx_main_container").append($cur);
        }
        $cur.css({
            "left": (pos.col * cel.w),
            "top": (pos.row * cel.h),
        });

	$(".appx-scroll-act-new").removeClass("appx-scroll-act-new");
	if( appxIsScrollReg(box) && !appxIsScrollAct(box) ) {
	    $("#box_" + box.widget.wBoxNumber.toString()).addClass("appx-scroll-act-new");
	}

        //TODO: can get this faster through query, see appxinvokehandler()
        for (var i = 0; i < box.items.length; i++) {
            var item = box.items[i],
                $tag = item[1];
            if (item[0].hasOwnProperty("pos_row")) { //Item, not box

                if (firstModItem == null && appxIsModifiable(item[0]))
                    firstModItem = item;

                //TODO: when we press left, the cursor is already outside the field and bCurItem test fails
                var pc = 0;
                if (appx_session.keyLeft) {
                    pc = 1;
                }

                if (
                    appxIsModifiable(item[0]) &&
                    pos.row >= item[0].pos_row &&
                    pos.row <= (item[0].pos_row + item[0].size_rows - 1) &&
                    pc + pos.col >= item[0].pos_col &&
                    pos.col <= (item[0].pos_col + item[0].size_cols - 1)
                ) {
                    bCurItem = true;
                    if (bFocus || true) {
			$tag.focus();
		    }
                    break;
                }
            }
        }

	if (appxIsLocked() && !bCurItem) {
	    var itmtag = null;
	    box = appxfindbox(pos.row, pos.col, 1, 1, false);
	    
	    for (var i = 0; i < box.items.length; i++) {
		var item = box.items[i];
		var $tag = item[1];

		if (item[0].hasOwnProperty("pos_row")) { //Item, not box		    
		    if (appxIsModifiable(item[0])) {
			itmtag = $tag;
		    }
		}
	    }

	    if (itmtag) {
		itmtag.focus();
	    }
	}

        if (firstModItem == null && (!bCurItem) && appx_session.topboxid != null) {
            for (var i = 0; i < appx_session.current_show.boxes[appx_session.topboxid].items.length; i++) {
                var item = appx_session.current_show.boxes[appx_session.topboxid].items[i],
                    $tag = item[1];
                if (item[0].hasOwnProperty("pos_row")) { //Item, not box
                    if (appxIsModifiable(item[0])) {
                        firstModItem = item;
                    }
                }
            }
        }

        if (appxIsLocked() && (!bCurItem) && firstModItem != null) {
            var $tag = firstModItem[1];
            bCurItem = true;
            $tag.focus();
	    focused = true;
        }

        $cur.css("visibility", ((bCurItem || appxIsLocked() || appx_session.getProp("drawBlockCursor") != "true") ? "hidden" : "visible"));
        return bCurItem;
    }
    catch (ex) {
        console.log("appxshowcursor: " + ex);
    }
}

function appxnestwidgets() {
    var boxNos = {};
    try {
        for (var i = 0; i < appx_session.current_show.boxes.length; i++) {
            var box = appx_session.current_show.boxes[i];
            box.widget = new Widget(0, "", box.data);
            if (box.widget) {
                boxNos[box.widget.wBoxNumber] = i;
            }
        }
    }
    catch (ex) {
        console.log("appxidboxes: " + ex);
    }

    try {
        // process items
        for (var i = 0; i < appx_session.items.length; i++) {
            var item = appx_session.items[i][0];
            var box = appxitembox(item.pos_row, item.pos_col, item.size_rows, item.size_cols);
            if (box) {
                box.items.push(appx_session.items[i]);
            }
        }
        appx_session.items = [];

        // process widgets
        for (var i = 0; i < appx_session.widgets.length; i++) {
            var wgt = appx_session.widgets[i];
            var wx = wgt[0];
            var boxno = wgt[0].boxid;
            var boxid = boxNos[boxno];
            var box = appx_session.current_show.boxes[boxid];
            if (!appxIsFullscreen(box) && appxIsScroll(box)) {
                var $tag = wgt[1];
                var scrollbox = box;
                var box = appxitembox(scrollbox.begin_row + wx.wPositionY - 1, wx.wPositionX, wx.wSizeH, wx.wSizeW);
                var topAdj = (box.begin_row - scrollbox.begin_row) + 1;
                var leftAdj = (box.begin_column - scrollbox.begin_column) + 1;
                $tag.css({
                    "top": (parseInt($tag.css("top")) - (topAdj * appx_session.rowheightpx)) + "px",
                    "left": (parseInt($tag.css("left")) - (leftAdj * appx_session.colwidthpx)) + "px"
                });
            }
            if (box) { box.widgets.push(appx_session.widgets[i]); }
            else { console.log("appxnestwidgets() failed to find box for widget"); }
        }
        appx_session.widgets = [];

        // process rowtext
        for (var i = 0; i < appx_session.rowtext.length; i++) {
            var rowtxt = appx_session.rowtext[i];
            var boxIdx; // index into boxes[] array.
            var boxNo; // Box number from Appx

            if (rowtxt.type == ROWTEXT_TYPE_ITEM) {
                box = appxitembox(rowtxt.pos_row, rowtxt.pos_col, rowtxt.size_rows, rowtxt.size_cols);
            }
            else {
                boxNo = rowtxt.boxid;
                boxIdx = boxNos[boxNo];
                if (boxIdx >= 0)
                    box = appx_session.current_show.boxes[boxIdx];
            }
            box.rowtext.push(rowtxt);
        }
        appx_session.rowtext = [];

    }
    catch (ex) {
        console.log("appxnestwidgets: " + ex);
    }
}

//Show Message Handler (CharView.updater?)
function appxshowhandler(x) {
    try {
        if (x == null || x.data == null) return;
        appx_session.current_show = (x.data);

        var show = appx_session.current_show;
        var mode = show.curraction[0];

        appxnestwidgets();
        appxshowboxes();
	setTimeout(function(){
		appxSetTabIndexes();
	    }, 0);

        screenflipped = false;
        if (appx_session.getProp("doubleBufferScreen") == false || appx_session.pendingResources.length == 0)
            appxshowscreen();

        if ((show.curraction[0] & M_SAVE) != 0) {
            appx_session.dirtySinceSave = false;
        }

        if ((show.curraction[0] & M_WAIT) == 0) {
            sendappxshow(0, []);
        }
        else {
            $(document).tooltip();
            blocked = false;
            try {
                $("#main").unblock();
            }
            catch (e) {}
            appxstarttimeout();
            if (appx_session.pendingResources.length == 0)
                appxSetStatusStateText(APPX_STATE_READY);
            else
                appxSetStatusStateText(APPX_STATE_IMAGES);
        }

    }
    catch (ex) {
        console.log("appxshowhandler: " + ex);
    }
    finally {
        appxSetLocked(false);
    }
}

function appxshowrowtextforbox(box) {
    var count = 0;
    var equalRow = 0;
    var headingOne = 0;
    var headingTwo = 0;

    while (box.rowtext.length > 0) {
        var rowtxt = box.rowtext.pop();
        var layer = 0;
        var boxNo = box.widget.wBoxNumber;
        if (rowtxt.pos_row == 1 && rowtxt.string.trim() == "Standard Toolbars")
            continue;
        if (!(appxIsScroll(box) || appxIsScrollReg(box))) {
            if (rowtxt.type == ROWTEXT_TYPE_ITEM) {
                rowtxt.pos_row -= box.begin_row - 1;
                rowtxt.pos_col -= box.begin_column - 1;
            }
        }
        if (rowtxt.string && rowtxt.string.length > 1 && rowtxt.string[0] == '=' && box.scrollrow && box.scrollrow == rowtxt.pos_row + 1) {
            equalRow = rowtxt.pos_row;
            continue;
        }
        if (equalRow > 0 && equalRow == rowtxt.pos_row + 1) {
            headingOne = rowtxt.pos_row;
            rowtxt.pos_row++;
        }
        if (equalRow > 0 && headingOne > 0 && equalRow == rowtxt.pos_row + 2) {
            headingTwo = rowtxt.pos_row;
            rowtxt.pos_row++;
        }
        if (rowtxt.pos_row > 0) {
            if (rowtxt.pos_row == 1 && box.widget && box.widget.wLabel && box.widget.wLabel.trim() == rowtxt.string.trim()) {
                continue;
            }
            count++;
            var x = 0,
                y = 0,
                w = 0,
                h = 0;
            var colW = appx_session.colwidthpx;
            var rowH = appx_session.rowheightpx;

            if (box && appxIsScrollReg(box)) {
                x = box.begin_column - 1;
                y = box.begin_row;
            }

            x = (rowtxt.pos_col - x) * colW;
            y = (rowtxt.pos_row - y) * rowH;
            w = rowtxt.size_cols * colW;
            h = rowtxt.size_rows * rowH;

            $tag = $("<div>");

            if (rowtxt.size_rows > 1)
                $tag.addClass("rowtextwrap");
            else
                $tag.addClass("rowtext");

            if (rowtxt.string.length > 1 && rowtxt.string[0] == '=')
                $tag.addClass("ColHdgSep");
            $tag.attr("id", "rowtext_" + count.toString());
            $tag.text(rowtxt.string.replace(/[\xb6\x0d]/g, "\n"));

            $tag.css({
                "margin": "0",
                "padding": "0"
            });

            var $box = $("<div>").addClass("appx--box"); //container
            //var clsText = TEXT_LEFT;
            //$box.addClass(clsText).appendTo($tag);
            $box.appendTo($tag);

            $tag.css({
                "position": "absolute",
                "left": x,
                "top": y,
                "height": h,
                "width": w,
                "z-index": -20
            });

            if (box != null) {
                var $box = $("#screenBuf #box_" + boxNo);
                if ($box.length) $box.append($tag);
            }
        }
    }
}

function appxshowitemsforbox(box, isActiveBox) {

    //sort array of boxes and items for tabindex
    for (var i = box.items.length - 1; i >= 0; i--) {
        var item = box.items[i][0];
	var itemNext = null;

	if( i > 0 )
	    itemNext = box.items[i-1][0];

        var boxtop = 0;

        clearBoxRowText(box, item.pos_row, item.pos_col, item.size_rows, item.size_cols);

        if (!appxIsScrollReg(box)) {
            boxtop = box.begin_row;
        }

        var $tag = box.items[i][1];

        if (isActiveBox) appxSetTabElement(item, $tag);

        if (item.pos_row - boxtop >= 0) {

            if (appxIsModifiableCapable(item) && appxIsScrollReg(box))
                $tag.removeClass("appx-not-modifiable");
            boxno = box.widget.wBoxNumber;

            appxwidgetdimensions(item, $tag, box);
            var $box = $("#screenBuf #box_" + boxno);
            if ($box.length) $box.append($tag);
            else console.log("no box...");

            //put a click handler on scrolling records
            if (appxIsScrollReg(box)) {
                //ignore a click on modifiable fields (inside the active record)
                if (!appxIsModifiable(item)) $tag.click(appxScrollClick);
            }
        }
        else {
            $_screenBuf.append($tag);
        }
        $tag.css("z-index", (box.layer) + 10);

        //<scan>
        if (appxIsScannable(item)) {
            var cel = appxGetCellSize();
	    var pad = 0;
	    var scanLeft = (item.pos_col * cel.w) + (item.size_cols * cel.w) + cel.w + pad;
	    var scanTop = (item.pos_row * cel.h) + 4;
	    var addScan = true;
	    debugger;
	    if( itemNext ) {
		var nextLeft = (itemNext.pos_col * cel.w);
		var nextTop = (item.pos_row * cel.h);
		if( item.pos_row == itemNext.pos_row ) {
		    if( scanLeft <= nextLeft && scanLeft + cel.w >= nextLeft )
			addScan = false;
		}
	    }

	    if( addScan ) {
		if( $tag.is("select") )
		    pad = cel.w * 2.5;
		
		$("<img>")
		    .attr("src", ""+appxClientRoot+"/images/scanicon.png")
		    .css({
			    "left": scanLeft,
				"position": "absolute",
				"top": scanTop,
				"z-index": $tag.css("z-index")
				})
		    .click({
			    col: item.pos_col,
				row: item.pos_row,
				tag: $tag
				}, function (e) {
			    appxPutCursor(e.data.col, e.data.row);
			    appxwidgetcallback(OPT_SCAN);
			})
		    .appendTo($_screenBuf);
	    }
        }
        //</scan>

    }
}

//double buffering
var $_screenBuf = null; //appx_session?
function appxshowscreen() { //flip
    //TODO: appx_main_container should become a flexible container id (e.g. new Appx('#mydiv'))
    var $screen = $("#appx_main_container");
    $screen.attr("id", "screenBuf");
    /*
    $screen.find(".ckeditor").each(function () {
	    var editor = CKEDITOR.instances[$(this).attr("name")];
	    editor.destroy();
    });
    */
    $screen.empty();
    $screen.hide();

    $_screenBuf.attr("id", "appx_main_container");
    $_screenBuf.show();
    $_screenBuf = $screen;

    createDropDownMenu();
    createPopupMenu();
    createdToolbarMenu();

    applymessages();
    appxshowcursor(true);
    screenflipped = true;
    applystyles();
    applytokens();

    if (appx_session.getProp("showScrollbar") == true) {
        var vcrcount = 0;
        $.each($(".appx-scroll"), function (i, el) {
            vcrcount++;
            var $grouphtml = $("<div id='appx-group-vcr-" + vcrcount + "'>").addClass("appx-vcr-group");
            var $vcrhtml = $("<div id='appx-scroll-vcr-" + vcrcount + "'>").addClass("appx-vcr").css({
                "position": "absolute",
                "z-index": 10000,
                "top": (parseInt($(el).css("top")) + 0) + "px",
                "left": (parseInt($(el).css("left")) + parseInt($(el).css("width")) - 16) + "px",
                "width": "16px",
                "height": $(el).css("height")
            });

            if (appx_session.getProp("dockingScrollbar") == true) {
                $grouphtml.addClass("appx-vcr-hide");
                $vcrhtml.hover(function () {
                    // mouse enter
                    $("#appx-group-vcr-" + vcrcount).removeClass("appx-vcr-hide");
                }, function () {
                    // mouse exit
                    $("#appx-group-vcr-" + vcrcount).addClass("appx-vcr-hide");
                });
            }

            var extra = Math.max(parseInt($(el).css("height")) - (16 * 6), 16) - 0;
            $grouphtml.html("");
            $grouphtml.append($("<button type='button'>").addClass("appx-vcr-btn").addClass("appx-vcr-up3").click(function () {
                event.preventDefault();
                appxwidgetcallback(OPT_SCROLL_FIRST);
                return false;
            }));
            $grouphtml.append($("<button type='button'>").addClass("appx-vcr-btn").addClass("appx-vcr-up2").click(function () {
                event.preventDefault();
                appxwidgetcallback(OPT_SCROLL_PREV);
                return false;
            }));
            $grouphtml.append($("<button type='button'>").addClass("appx-vcr-btn").addClass("appx-vcr-up1").click(function () {
                event.preventDefault();
                appxwidgetcallback(OPT_SCROLL_DOWN);
                return false;
            }));
            $grouphtml.append($("<button type='button'>").addClass("appx-vcr-btn").addClass("appx-vcr-scan").css({
                "height": extra + "px"
            }).click(function () {
                event.preventDefault();
                appxwidgetcallback(OPT_KEY_ENTRY);
                return false;
            }));
            $grouphtml.append($("<button type='button'>").addClass("appx-vcr-btn").addClass("appx-vcr-down1").click(function () {
                event.preventDefault();
                appxwidgetcallback(OPT_SCROLL_UP);
                return false;
            }));
            $grouphtml.append($("<button type='button'>").addClass("appx-vcr-btn").addClass("appx-vcr-down2").click(function () {
                event.preventDefault();
                appxwidgetcallback(OPT_SCROLL_NXT);
                return false;
            }));
            $grouphtml.append($("<button type='button'>").addClass("appx-vcr-btn").addClass("appx-vcr-down3").click(function () {
                event.preventDefault();
                appxwidgetcallback(OPT_SCROLL_LAST);
                return false;
            }));

            $vcrhtml.html($grouphtml);
            $("#appx_main_container").append($vcrhtml);
        });
    }
}

function intersects(left1, right1, top1, bottom1, left2, right2, top2, bottom2) {
    return !(left2 > right1 ||
        right2 < left1 ||
        top2 > bottom1 ||
        bottom2 < top1);
}

var spaces = "                                                  " +
    "                                                  " +
    "                                                  " +
    "                                                  " +
    "                                                  ";

function clearBoxRowText(box, r, c, h, w) {
    for (var ii = 0; ii < box.rowtext.length; ii++) {
        var rowtxt = box.rowtext[ii];
        if (intersects(rowtxt.pos_col, rowtxt.pos_col + rowtxt.size_cols - 1, rowtxt.pos_row, rowtxt.pos_row + rowtxt.size_rows - 1, c, c + w - 1, r, r + h - 1)) {
            if (rowtxt.pos_row == 1 && rowtxt.string.trim() == "Standard Toolbars")
                continue;
            var colStart = Math.max(rowtxt.pos_col, c) - rowtxt.pos_col;
            var colEnd = Math.min(rowtxt.pos_col + rowtxt.size_cols, c + w) - rowtxt.pos_col;
            var newStr = (colStart > 0 ? rowtxt.string.substring(0, colStart) : "") +
                (spaces.substring(0, colEnd - colStart)) +
                (colEnd < rowtxt.string.length ? rowtxt.string.substring(colEnd) : "");
            rowtxt.string = newStr;
        }
    }
}

function appxshowwidgetsforbox(box, isActiveBox) {
    while (box.widgets.length > 0) {
        var wgt = box.widgets.shift(),
            $tag = wgt[1];
        var wx = wgt[0];
        var boxno = box.widget.wBoxNumber;

        if (parseInt(wx.wWidgetType) == WIDGET_TYPE_ROWTEXT && wx.wLabel[5] == '=')
            continue;

        if (parseInt(wx.wWidgetType) != WIDGET_TYPE_BOX)
            clearBoxRowText(box, wx.wPositionY, wx.wPositionX, wx.wSizeH, wx.wSizeW);

        $("#screenBuf #box_" + boxno).append($tag);
        if (wgt[0].wMovable) {
            $tag.data({
                "wMovableCommand": wgt[0].wMovableCommand
            })
            $tag.draggable({
                cancel: false,
                stop: function () {
                    var r = Math.floor(($(this).offset().top - $("#appx_main_container").offset().top) / appx_session.rowheightpx);
                    var c = Math.floor(($(this).offset().left - $("#appx_main_container").offset().left) / appx_session.colwidthpx);
                    appxPutCursor(c, r);

                    if ($(this).data("wMovableCommand")) {
                        appxwidgetcallback($(this).data("wMovableCommand"));
                    }
                }
            });
        }

        if (isActiveBox && parseInt(wgt[0].wWidgetType) == WIDGET_TYPE_BUTTON) {
            appxSetTabElement(wx, $tag);

        }
        else {
            //chrome makes fieldsets selectable when they have a tabindex
            if (!$tag.is("fieldset"))
                $tag.attr("tabindex", -1);
            if (parseInt(wgt[0].wWidgetType) == WIDGET_TYPE_BUTTON) {
                $tag.prop("disabled", true);
                $tag.css({
                    "color": ""
                });
            }
            $tag.removeClass("default"); //can be a button on an underlying frame
        }

    }
}

/* (CharView.updater)
20141112_ca_taborder&autotabout
20141218_ca_#3250_tabgroups

Help (Tab Level) 0-255
Defines the physical pass over the whole screen with respect to tab order. 
If left blank, this is considered tab level 0.
APPX will tab from top-left to bottom-right for all fields in tab level 0.
Then repeat for all buttons in tab level 0.
This cycle will then repeat for each tab level found.
This behavior can be altered by the specification of Tab Groups.

Help (Tab Group) X(4)
Tab Groups can be used to mark related fields and buttons so that they will be grouped together as part of the tab order.
While making a pass over the screen for a "Tab Level", if a field is hit which has a non-null "Tab Group",
the tab order will then proceed to all the fields and buttons that are coded to that Tab Level/Tab Group.
After reaching the last field or button in the group, the order will continue where it left off prior to 
entering the group.
Note that for "Tab Groups", fields and buttons are consolidated into the same top-left to bottom-right pass.
*/
/** appxSetTabElement
 * called from appxshowitemsforbox and appxshowwidgetsforbox
 * expands a list to be processed by appxSetTabIndexes looking like this:
 *
 * [ Levels ]
 * 0
 *     [ Elements ]
 *     0 { grp, tag } {*}
 *     1..
 *     [ Groups ] (string key)
 *     A
 *         [ Elements ]
 *         0 { grp, tag }
 *         1..
 *     B..
 * 1..
 *
 * {*} The first encounter of a non-empty grp is stored in top [Elements], then added
 *     to [Groups] with other members, so they can be placed on this spot later.
 */
function appxSetTabElement(wori, $tag) {
    try {
        if (wori && $tag) {

            var saveTabElement = function (wdgt) {
                if (!wdgt || wdgt.wEnabled != false) {
                    // initialize tablist to array when needed, set to null in appxshowboxes
                    var tab = appx_session.tablist = appx_session.tablist || [];
                    var lvl = (wdgt ? parseInt(wdgt.wTabGroup) : 0);
                    var grp = (wdgt ? wdgt.wTabSubGroupId || '' : '');
                    var obj = {
                        'grp': grp,
                        'tag': $tag
                    };
                    if (isNaN(lvl)) lvl = 0;
                    tab[lvl] = tab[lvl] || [];
                    if (grp != '') {
                        if (!tab[lvl][grp]) {
                            // first group element, add to our normal level pass list
                            // init array for elements in this group and level
                            tab[lvl].push(obj);
                            tab[lvl][grp] = [];
                        }
                        tab[lvl][grp] = tab[lvl][grp] || [];
                        tab[lvl][grp].push(obj);
                    }
                    else {
                        tab[lvl].push(obj);
                    }
                }
            };

            //TODO: add function(s) appxIsItem and/or appxIsWidget
            if (wori.hasOwnProperty("wWidgetType")) { //button widget?
                if (parseInt(wori.wWidgetType) == WIDGET_TYPE_BUTTON) {
                    saveTabElement(wori);
                }
            }
            else if (wori.hasOwnProperty("widget")) { //modifiable item?
                if (appxIsModifiable(wori)) {
                    if (!$tag.prop("disabled")) {
                        saveTabElement(wori.widget);
                    }
                }
            }
        }
    }
    catch (ex) {
        console.log("appxSetTabElement: " + ex);
    }
}

// called from tag keyhandlers attached in appxSetTabIndexes
function appxSetTabFocus($tagSrc, bBack) {
    try {
        var ti = parseInt($tagSrc.attr("tabindex"));
        var tl = appx_session.tablist;

	var el = tl[ti];
	if (el.hasClass("appxdatefield") && el.find(".appxdatevalue"))
	    el = el.find(".appxdatevalue");

        if (bBack) 
	    el.blur();

        ti = (bBack ? --ti : ++ti);

        if (ti < 0) 
	    ti = tl.length - 1;
        else if (ti >= tl.length) 
	    ti = 0;

	el = tl[ti];
	if (el.hasClass("appxdatefield") && el.find(".appxdatevalue"))
	    el = el.find(".appxdatevalue");
	
        el.focus(); //always for buttons

        if (appx_session.getProp("autoSelect")) {
            el.select();
        }
    }
    catch (ex) {
        console.log("appxSetTabFocus: " + ex);
    }
}

// called from appxshowhandler, sort and apply tabable buttons and items
function appxSetTabIndexes() {
    try {

        if (appx_session.tablist) {

            var tablist = [];

            var pushTabElement = function ($tag) {
                tablist.push($tag);
                $tag.attr("tabindex", tablist.length - 1); //unique sorted tabindex
                $tag.on("keydown", function (ke) {
                    if (ke.which == 9) { //tab
                        appxSetTabFocus($(this), ke.shiftKey);
                    }
                    else if (ke.which == 13 && $(this).is("button")) { //enter
                        $(this).trigger("click");
                    }
                    else { //cancel input when communicating (e.g. key pause)
                        if (!appxIsLocked()) return true;
                    }
                    ke.preventDefault();
                    ke.stopPropagation();
                    return false;
                });
                if (!$tag.is("button")) {
                    // Implements Auto Tab-Out 
                    // modifier and non-printing keys such as Shift, Esc, Del 
                    // trigger keydown events and not keypress events
                    // using keyup however, because keypress sometimes doesn't
                    // overwrite chars after a select
                    $tag.on("keyup", function (ke) {
                        try {
                            if (!ke.altKey && !ke.ctrlKey && !ke.metaKey) {
                                var k = ke.which;
                                //http://www.cambiaresearch.com/articles/15/javascript-char-codes-key-codes
                                //skip nonprintable keycodes
                                if (k > 46 && (k < 91 || k > 93) && (k < 112 || k > 125)) {
                                    var max = $(this).attr("maxlength");
                                    if (max && max <= $(this).val().trim().length) {
                                        if (appx_session.getProp("autoTabOut")) {
                                            appxSetTabFocus($(this)); //select next item
                                        }
                                    }
                                }
                            }
                        }
                        catch (ex) {
                            alert("$tag keypress: " + ex);
                        }
                        return true; //bubble
                    });
                }
            }; //pushTabElement

            // TODO? should we check by rows and columns due to micro-positioning?
            var sortTabElements = function (objA, objB) {
                var posA = objA.tag.position();
                var posB = objB.tag.position();
		posA.top = parseInt(objA.tag.css("top"));
		posA.left = parseInt(objA.tag.css("left"));
		posB.top = parseInt(objB.tag.css("top"));
		posB.left = parseInt(objB.tag.css("left"));

                if (posA.top < posB.top) {
                    return -1; // lower 'row', put before
                }
                else if (posA.top > posB.top) {
                    return 1; // higher 'row', put after
                }
                if (posA.left < posB.left) { //same top, check left
                    return -1;
                }
                else if (posA.left > posB.left) {
                    return 1;
                }
                return 0; //same top and left
            };
            var tab = appx_session.tablist;
            for (var lvl = 0; lvl < tab.length; lvl++) {
                // tab[lvl] could be undefined, because unused indexes in between
                // get allocated too, for example if we have levels 0 and 2,
                // tab will be an array containing 0, 1 (undefined) and 2
                // not bothering since it is a volatile list and has a maximum of 255
                // (using a string index would need sorting and for .. in ..) 
                if (tab[lvl]) {
                    tab[lvl].sort(function (objA, objB) {
                        // items go in front of buttons inside a tab level
                        // tab groups don't care about element types
                        if (!objA.tag.is("button") && objB.tag.is("button")) {
                            return -1;
                        }
                        if (objA.tag.is("button") && !objB.tag.is("button")) {
                            return 1;
                        }
                        // both element types are equal, check positions
                        return sortTabElements(objA, objB);
                    });
                    // i: elements belonging to this tab level
                    for (var i = 0; i < tab[lvl].length; i++) {
                        var obj = tab[lvl][i];
                        // could encounter the first element of a group
                        if (obj.grp != '') {
                            if (tab[lvl][obj.grp]) {
                                tab[lvl][obj.grp].sort(sortTabElements);
                                // j: each item or button belonging to this group
                                for (var j = 0; j < tab[lvl][obj.grp].length; j++) {
                                    pushTabElement(tab[lvl][obj.grp][j].tag);
                                }
                                tab[lvl][obj.grp] = null; //done with it
                            }
                        }
                        else {
                            pushTabElement(obj.tag);
                        }
                    }

                }
            }

            // apply list filled through pushTabElement and used by appxSetTabFocus
            appx_session.tablist = tablist;
        }

    }
    catch (ex) {
        console.log("appxSetTabIndexes: " + ex);
    }
}

//Sends show data to the server
/* TShow -> MDisplay
  public byte[] terminal = new byte[4];  0
  public byte mode; 0 0 0                4
  public byte[] keymap = new byte[4];    8
  public MRCBlock cursor;               12 (+8) row, col = 16
  public int timeout;                   20
  public byte[] charId = new byte[4];   24
  public int option;                    28
  public int status;                    32
  public int boxCount;                  36
  public byte[] boxPtr = new byte[4];   40
    public MRCBlock cursorAlt;          44 (+8)
    public int optionAlt;               52
 */
function sendappxshow(option, data) {
    appxSetStatusStateText(APPX_STATE_BUSY);
    try {

        $(".appxtablewidget").each(function () {
            var uigrid = $("#newtablewidget_" + $(this).attr('id'));
            var selrows = uigrid.getGridParam("selarrrow");
	    var sortcol = uigrid.getGridParam("sortname");
	    var sortorder = uigrid.getGridParam("sortorder");
            var uigridbdiv = $(this).find(".ui-jqgrid-bdiv");
            if (uigridbdiv) {
                var gridprops = null;

                if (!gridprops) {
                    gridprops = new GridCacheProps();
                }

                gridprops.scrollTop = uigridbdiv.scrollTop();
                gridprops.scrollLeft = uigridbdiv.scrollLeft();
                gridprops.selrows = selrows;
		gridprops.sortcol = sortcol;
		gridprops.sortorder = sortorder;
                appx_session.gridpropscache[$(this).data("tableHashKey")] = gridprops;
            }

	    var rows = [];
	    for (var i = 0; i < selrows.length; i++) {
		rows.push(uigrid.getRowData(selrows[i]).id2);
	    }
	    $(this).data("selkeys", rows);

        });

        //TODO? is the option 0 check needed?
        if (appxIsLocked() && option != 0 && option != 0xFFFF) return; //e.g. obsolete focus handler
        appx_session.applyStylesCount = 0;
        appx_session.currenttabledata = [];
        appxSetLocked(true);
        appxstoptimeout();

        appx_session.override = false;

        var showstructints = [];
        var rtnshow = appx_session.current_show;
        if (rtnshow != null) {

            var mode = rtnshow.rawdata[4];

            //have to set return status in 0xC message to indicate and option was selected
            if (option >= 0) rtnshow.rawdata[35] = 1;

            var dv = new DataView(new Uint8Array(rtnshow.rawdata).buffer);

            //20140314 chris@praclox.nl cursor
            var cur = appxGetCursorPos();
            dv.setUint32(12, appxFixCursorRow(cur.row));
            dv.setUint32(16, appxFixCursorCol(cur.col));
            dv.setUint32(28, parseInt(option));

            if (mode == 2 || mode == 32) {
		dv.setUint32(32, parseInt("8"));
		dv.setUint32(28, parseInt("65535"));
	    }

            dv.setInt32(44, 0);
            dv.setInt32(48, 0);
            dv.setUint32(52, 0);

            if (rtnshow.altuseroption) { // swap current cusrsor and option with alt cursor and option

                var temp1 = 0;
                var temp2 = 0;

                temp1 = dv.getUint32(12);
                temp2 = new DataView(new Uint8Array(rtnshow.altcurserrow).buffer).getUint32(0);
                dv.setUint32(12, temp2);
                dv.setUint32(44, temp1);

                temp1 = dv.getUint32(16);
                temp2 = new DataView(new Uint8Array(rtnshow.altcursorcol).buffer).getUint32(0);
                dv.setUint32(16, temp2);
                dv.setUint32(48, temp1);

                rtnshow.altuseroption = null;
                rtnshow.altcurserrow = null;
                rtnshow.altcursorcol = null;;
            }

            var showstructints = [];
            for (var i = 0; i < new Uint8Array(dv.buffer).length; i++) {
                showstructints[i] = dv.getUint8(i);
            }
        }

        var ms = {
            cmd: 'appxmessage',
            args: [0, 0, 0, 2, 12, 0, 0, 0],
            handler: 'appxshowhandler',
            data: null
        };
        appx_session.ws.send(JSON.stringify(ms));

        ms = {
            cmd: 'appxmessage',
            // we need to calculate the screen size here based on attach function(defaults to 31x110 which require 54 bytes to send repeating (ff 00)s which means ((255 x (54/2))/2 = 3410
            args: [0, 54, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 209, 0],
            handler: 'appxshowhandler',
            data: null
        };
        appx_session.ws.send(JSON.stringify(ms));

        ms = {
            cmd: 'appxmessage',
            args: showstructints,
            handler: 'appxshowhandler',
            data: null
        };
        appx_session.ws.send(JSON.stringify(ms));

        if ((mode & M_COMPARE) != 0 && appx_session.disrtySinceSave ) { //  || data.length > 0) {
            ms = {
                cmd: 'appxmessage',
                args: [1],
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));
        }
        else {
            ms = {
                cmd: 'appxmessage',
                args: [0],
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));
        }
        if (appx_session.override) {

            ms = {
                cmd: 'appxmessage',
                args: [0, 0, 0, 2, 86, 0, 0, 0, 0, 0, 0, 1, 11, 27, 0, 3, 84, 82, 75],
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

        }
        else if (data.length > 0) {

            ms = {
                cmd: 'appxmessage',
                args: [0, 0, 0, 2, 86, 0, 0, 0],
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            var msgdata = [];
            for (var dl = 0; dl < data.length; dl++) {
                var dataval = "";
                // We should add a handler in keydown event to change to upper if css text tranform UPPER case is set
                // Or check field class here to see if it should be uppercase
		if ($(data[dl]).find(".appxdatevalue").val()) 
		    dataval = $(data[dl]).find(".appxdatevalue").val();
		else if ($(data[dl]).closest(".appxitem").find("input").val())
		    dataval = $(data[dl]).closest(".appxitem").find("input").val();
		else if ($(data[dl]).val())
		    dataval = $(data[dl]).val();
		else if ($(data[dl]).closest(".appxitem").val()) 
		    dataval = $(data[dl]).closest(".appxitem").val();
		else if ($(data[dl]).find(".appx-data").html()) 
		    dataval = $(data[dl]).find(".appx-data").html();
		
                if ($(data[dl]).css("text-transform") == "uppercase") 
                    dataval = dataval.toUpperCase();

                var id = $(data[dl]).closest(".appxitem").attr('id');
		if( id == "" || !id ) {
		    id = $(data[dl]).closest(".appxitem").closest("div").attr('id');
		}
                var ar = getClientId(id).split('_');
                var datarow = parseInt(ar[2]);
                msgdata.push(datarow);
                var datacol = parseInt(ar[1]);
                msgdata.push(datacol);

                var u8strarray = toUTF8Array(dataval);
		var byte2 = u8strarray.length % 256;
		var byte1 = (u8strarray.length - byte2) / 256;
                msgdata.push(byte1);
                msgdata.push(byte2);
                msgdata = msgdata.concat(u8strarray);
            }

	    clearClientIds();

            var msglength = [0, 0, 0];
            msglength.push(data.length);

            ms = {
                cmd: 'appxmessage',
                args: msglength,
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            ms = {
                cmd: 'appxmessage',
                args: msgdata,
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

        }
        else {

            ms = {
                cmd: 'appxmessage',
                args: [0, 0, 0, 2, 86, 0, 0, 0, 0, 0, 0, 0],
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

        }

        //END SENDSHOW and LONGDATA

        /* BEGIN-------------------------------- */
        /* SEND TABLE DATA - ALWAYS SEND MESSAGE */
        /* even when length of table data is 0   */
        /* ------------------------------------- */

        ms = {
            cmd: 'appxmessage',
            args: [0, 0, 0, 4, 90, 0, 0, 0],
            handler: 'appxshowhandler',
            data: null
        };
        appx_session.ws.send(JSON.stringify(ms));

        var keycnt = 0;
        data = $(".appxtablewidget");
        var msgdata = [];
        for (var dl = 0; dl < data.length; dl++) {

            var dataobj = $(data[dl]).data();

            if (dataobj.selkeys) {
                for (var z = 0; z < dataobj.selkeys.length; z++) {
                    var dataval = "";
                    // We should add a handler in keydown event to change to upper if css text tranform UPPER case is set
                    // Or check field class here to see if it should be uppercase
                    dataval = dataobj.selkeys[z];

                    //ROW - need 4bytes BE
                    var datarow = [0, 0, 0, parseInt(dataobj.row)];
                    msgdata = msgdata.concat(datarow);
                    //COL - need 4bytes BE
                    var datacol = [0, 0, 0, parseInt(dataobj.col)];
                    msgdata = msgdata.concat(datacol);
                    //SEQ - need 4bytes BE
                    var seq = [0, 0, 0, keycnt + 1];
                    msgdata = msgdata.concat(seq);
                    //alway a 6
                    msgdata.push(6);

                    msgdata.push(dataval.length / 2);
                    for (var k = 0; k < dataval.length; k += 2) {
                        msgdata.push(parseInt(dataval.substring(k, k + 2), 16));
                    }
                    keycnt++;
                }
            }
        }

        //RUN THRU 4 BYTE BE
        var msglength = [0, 0, 0];
        msglength.push(keycnt);

        ms = {
            cmd: 'appxmessage',
            args: msglength,
            handler: 'appxshowhandler',
            data: null
        };

        appx_session.ws.send(JSON.stringify(ms));

        if (data.length > 0) {
            ms = {
                cmd: 'appxmessage',
                args: msgdata,
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));
        }

        /* END---------------------------------- */
        /* SEND TABLE DATA - ALWAYS SEND MESSAGE */
        /* even when length of table data is 0   */
        /* ------------------------------------- */

        //<dnd>
        /* BEGIN-------------------------------- */
        /* SEND DROP DATA  - ALWAYS SEND MESSAGE */
        /* even when length of table data is 0   */
        /* ------------------------------------- */

        /*
         **  DropInfo Data is organized in this way...
         **
         **  HEADER - 8 byte standard header with a data length of 4 and type 92
         **           where the 4 bytes of data is a uint32 count of dropinfo
         **           records to follow.
         **  count  - uint32 count of dropinfo records to follow.
         **
         **  ===== DropInfo Record, one per count =====
         **
         **    rc_blk - uint32 row, uint32 col of the widget that was dropped onto.
         **    parent - uint32 of the widget parent type of the widget dropped onto.
         **    path   - uint16 string length of path followed by string bytes.
         **    name   - uint16 string length of name followed by string bytes.
         **    ext    - uint16 string length of extention followed by string bytes.
         **    type   - uint16 string length of type followed by string bytes.
         **    size   - uint32 file size
         **    pcount - uint16 count of properties to follow for this dropinfo record.
         **
         **  ===== Property Record, one per pcount =====
         **
         **      keyword - uint16 string length of keyword followed by string bytes.
         **      value   - uint16 string length of value followed by string bytes.
         */

        ms = {
            cmd: 'appxmessage',
            args: [0, 0, 0, 4, 92, 0, 0, 0],
            handler: 'appxshowhandler',
            data: null
        };
        appx_session.ws.send(JSON.stringify(ms));

        var dropcount = appx_session.dndData.length;
        ms = {
            cmd: 'appxmessage',
            args: hton32(dropcount),
            handler: 'appxshowhandler',
            data: null
        };
        appx_session.ws.send(JSON.stringify(ms));

        for (var loop = 0; loop < dropcount; loop++) {

            var mDnD = appx_session.dndData[loop];

            //Should be a uint RC_BLK of the widget position
            //4 byte row
            ms = {
                cmd: 'appxmessage',
                args: hton32(mDnD.row),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));
            //4 byte col
            ms = {
                cmd: 'appxmessage',
                args: hton32(mDnD.col),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // 4 byte integer parent type
            ms = {
                cmd: 'appxmessage',
                args: hton32(mDnD.parentType),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // 2 byte integer dropinfo path len
            ms = {
                cmd: 'appxmessage',
                args: hton16(mDnD.path.length),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // string dropinfo path
            ms = {
                cmd: 'appxmessage',
                args: toUTF8Array(mDnD.path),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // 2 byte integer dropinfo name len
            ms = {
                cmd: 'appxmessage',
                args: hton16(mDnD.name.length),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // string dropinfo name
            ms = {
                cmd: 'appxmessage',
                args: toUTF8Array(mDnD.name),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // 2 byte integer dropinfo ext len
            ms = {
                cmd: 'appxmessage',
                args: hton16(mDnD.ext.length),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // string dropinfo ext
            ms = {
                cmd: 'appxmessage',
                args: toUTF8Array(mDnD.ext),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // 2 byte integer dropinfo type len
            ms = {
                cmd: 'appxmessage',
                args: hton16(mDnD.mtype.length),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // string dropinfo type
            ms = {
                cmd: 'appxmessage',
                args: toUTF8Array(mDnD.mtype),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // 4 byte int size (this is where we really need more than cb)
            ms = {
                cmd: 'appxmessage',
                args: hton32(mDnD.size),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));

            // 2 byte integer propcount
            var propcount = mDnD.props.length;
            ms = {
                cmd: 'appxmessage',
                args: hton16(propcount),
                handler: 'appxshowhandler',
                data: null
            };
            appx_session.ws.send(JSON.stringify(ms));
            for (var proploop = 0; proploop < propcount; proploop++) {
                var prop = mDnD.props[proploop];

                // 2 byte integer dropinfo keyword len
                ms = {
                    cmd: 'appxmessage',
                    args: hton16(prop.name.length),
                    handler: 'appxshowhandler',
                    data: null
                };
                appx_session.ws.send(JSON.stringify(ms));

                // string dropinfo keyword
                ms = {
                    cmd: 'appxmessage',
                    args: toUTF8Array(prop.name),
                    handler: 'appxshowhandler',
                    data: null
                };
                appx_session.ws.send(JSON.stringify(ms));

                // 2 byte integer dropinfo value len
                ms = {
                    cmd: 'appxmessage',
                    args: hton16(prop.value.length),
                    handler: 'appxshowhandler',
                    data: null
                };
                appx_session.ws.send(JSON.stringify(ms));

                // string dropinfo value
                ms = {
                    cmd: 'appxmessage',
                    args: toUTF8Array(prop.value),
                    handler: 'appxshowhandler',
                    data: null
                };
                appx_session.ws.send(JSON.stringify(ms));
            }
        }
        appx_session.dndData = [];
        //</dnd>

    }
    catch (ex) {
        console.log("sendappxshow: " + ex);
    }

}

//TShow
//var M_CHECK   = 0x01;
var M_SHOW = 0x02;
//var M_ALARM   = 0x04;
var M_WAIT = 0x08;
//var M_PAINT   = 0x10;
var M_COMPARE = 0x20;
var M_SAVE    = 0x40;
//MBox
var SCROLL = 0x00010000; //Scrolling area
var SCROLL_REG = 0x00020000; //Scrolling Region
var SCROLL_ACT = 0x00040000; //Scrolling ActiveReg
var IN_PROGRESS = 0x00080000; // box is in progress message
//var BORDER      = 0x08000000;
//Document long constants
/*var CHR_SCROLL =     0x0000000000010000;//Scrolling area
var CHR_SCROLL_REG = 0x0000000000020000;//Scrolling Region
var CHR_SCROLL_ACT = 0x0000000000040000;//Scrolling ActiveReg
var CHR_BORD_BOX = 0x0000000008000000;//complete rect border*/

function appxIsLocked() {
    return appx_session.locked;
}

function appxSetLocked(b) {
    appx_session.locked = b;
}

function appxGetCellSize() {
    return {
        "w": appx_session.colwidthpx,
        "h": appx_session.rowheightpx
    };
}

//clamp column
function appxFixCursorCol(c) {
    if (c < 1) c = 1;
    if (c > appx_session.screencols) c = appx_session.screencols;
    if (c > 255) c = 255;
    return c;
}

//clamp line/row
function appxFixCursorRow(r) {
    if (r < 1) r = 1;
    if (r > appx_session.screenrows - 3) r = appx_session.screenrows - 3;
    if (r > 255) r = 255;
    return r;
}

/** Display.getCursorPos / Document.lookupCursorPosition
 * Get the current cursor position.
 */
function appxGetCursorPos() {
    var c = appxFixCursorCol(ab2int32(appx_session.current_show.cursorcol));
    var r = appxFixCursorRow(ab2int32(appx_session.current_show.curserrow));
    return {
        "col": c,
        "row": r
    };
}

function appxIsInProgress(box) {
    return ((box.bit_mask & IN_PROGRESS) != 0 && (box.widget == null || box.widget.wWidgetType != 203));
}

function appxIsEmBuild(box) {
    return ((box.bit_mask & IN_PROGRESS) != 0 && (box.widget != null && box.widget.wWidgetType == 203));
}

function appxIsFullscreen(box) {try{
    return (
            (box.end_column - box.begin_column) + 1 >= appx_session.screencols &&
            (box.end_row - box.begin_row) + 1 >= (appx_session.screenrows - 3)
    );
}catch(ex){console.log("screen.appxIsFullscreen: " + ex);}}

function appxIsScroll(box) {
    return ((box.bit_mask & SCROLL) != 0);
}

function appxIsScrollAct(box) {
    return ((box.bit_mask & SCROLL_ACT) != 0);
}

function appxIsScrollReg(box) {
    return ((box.bit_mask & SCROLL_REG) != 0);
}

/** Display.putCursor / Document.setCursorPos
 * Puts the cursor at the specified position.
 * @param c column
 * @param r row //l line
 */
function appxPutCursor(c, r) {
    try {
        var bCurItem = false; //cursor inside field?
        if (!appxIsLocked()) {
            var cur = appxGetCursorPos(); //prevent recursive call through focus
            if (cur.col != c || cur.row != r) {
                appx_session.current_show.cursorcol = [0, 0, 0, appxFixCursorCol(c)];
                appx_session.current_show.curserrow = [0, 0, 0, appxFixCursorRow(r)];
                bCurItem = appxshowcursor(true); //don't call focus
            }
        }
        return bCurItem;
    }
    catch (ex) {
        console.log("appxPutCursor: " + ex);
    }
}

function appxPutTabOut() {
    try {
        if (!appxIsLocked()) {
            var cur = appxGetCursorPos();
            appx_session.current_show.altcursorcol = [0, 0, 0, appxFixCursorCol(cur.col)];
            appx_session.current_show.altcurserrow = [0, 0, 0, appxFixCursorRow(cur.row)];
            appx_session.current_show.altuseroption = [0, 0, 1, 76];
        }
    }
    catch (ex) {
        console.log("appxPutTabOut: " + ex);
    }
}

function appxScrollClick() {
    var col = $(this).data("col");
    var row = $(this).data("row");
    if (!col || !row) {
        return;
    }
    appxPutCursor(col, row);
    appxwidgetcallback(OPT_ENTER);
}

function appxSetStatusDbText(str) {
    $("#appx_status_db").html(str);
}

function appxSetStatusApText(str) {
    $("#appx_status_ap").html(str);
}

function appxSetStatusVerText(str) {
    $("#appx_status_ver").html(str);
}

function appxSetStatusUserText(str) {
    $("#appx_status_user").html(str);
}

function appxSetStatusDateText(str) {
    $("#appx_status_date").html(str);
}

function appxSetStatusKeymapText(str) {
    $("#appx_status_keymap").html(str);
}

function appxSetStatusMsgText(str) {
    appxSetStatusText(str);
}

function appxSetStatusModeText(str) {
    $("#appx_status_mode").html(str);
}

var APPX_STATE_BUSY = 0;
var APPX_STATE_READY = 1;
var APPX_STATE_DIRTY = 2;
var APPX_STATE_IMAGES = 3;
var APPX_STATE_EMS = 4;
var appx_state_last = -1;

function appxSetStatusStateText(state) {
    appx_state_last = state;

    var $str = $("<span>");

    switch (state) {
        case APPX_STATE_BUSY:
            $str.html("busy").addClass("appx-state-busy");
            break;
        case APPX_STATE_READY:
            $str.html("ready").addClass("appx-state-ready");
            break;
        case APPX_STATE_DIRTY:
            $str.html("ready+").addClass("appx-state-dirty");
	    appx_session.dirtySinceSave = true;
            break;
        case APPX_STATE_IMAGES:
            $str.html("images").addClass("appx-state-images");
            break;
        case APPX_STATE_EMS:
            $str.html("compile").addClass("appx-state-ems");
            break;
    }

    $("#appx_status_stat").html($str);
}

function appxSetStatusEmbldText(str) {
    $("#appx_status_embld").html(str);
}

function appxSetStatusProgressText(str) {
    $("#appx_status_progress").html(str);
}

function appxClearStatusMsgText() {
    $("#appx-status-msg").html("");
}

function appxSetStatusText(str) {
    str = str.trim();
    var $bar = $("#appx-status-msg");
    if ($bar.length > 0) {
        var clr = (str == "" ? "white" : "lime");
        if (str != "") {
            if (str.indexOf("Error - ") != -1) clr = "red";
            else if (str.indexOf("Warning - ") != -1) clr = "yellow";
        }
        if (str != "") {
            $bar.prepend(str + "\n\n");
        }
    }
}

function appxstarttimeout() {
    appxstoptimeout();
    var s = appx_session.current_show;
    if (s.timeout == null || s.timeout <= 0 || s.timeout > 60000) return;
    appx_session.interactT = setTimeout(function () {
        appxwidgetcallback(s.useroption ? ab2int32(s.useroption) : OPT_DIR_PROC_1);
    }, (s.timeout * 1000));
}

function appxstoptimeout() {
    if (appx_session.interactT) clearTimeout(appx_session.interactT);
    appx_session.interactT = null;
}

function appxstartblurtimeout() {
    appxstopblurtimeout();
    appx_session.interactTblur = setTimeout(function () {
        appxwidgetcallback(OPT_TAB_OUT);
    }, (1000));
}

function appxstopblurtimeout() {
    if (appx_session.interactTblur) clearTimeout(appx_session.interactTblur);
    appx_session.interactTblur = null;
}

function RowTextStruct() {
    var self = this;
    this.type = ROWTEXT_TYPE_NONE;
    this.boxid = 0;
    this.pos_row = 0;
    this.pos_col = 0;
    this.size_rows = 0;
    this.size_cols = 0;
    this.string = null;
};

var ROWTEXT_TYPE_NONE = 0;
var ROWTEXT_TYPE_ITEM = 1;
var ROWTEXT_TYPE_WIDGET = 2;

function createAppxGrid(elem) {
    var sizeH = 21;
    var sizeW = 8;
    var ratioW = Math.floor($(window).width()/sizeW),
        ratioH = Math.floor($(window).height()/sizeH);
    
    var parent = $('<div />', {
	    class: 'grid', 
	    width: ratioW  * sizeW, 
	    height: ratioH  * sizeH
	}).addClass('grid').appendTo(elem);

    for (var i = 0; i < ratioH; i++) {
        for(var p = 0; p < ratioW; p++){
            $('<div />', {
		    width: sizeW - 1, 
			height: sizeH - 1
			}).appendTo(parent);
        }
    }
}

