﻿//  *********************************************************
//  Configuration Section - Begin
//  *********************************************************

var serverConnectorVersionStr = "5.4.2";
var serverConnectorVersionNum = 50402.013;

var connectorPort = 3014;

var sslEnabled       = false;
var cryptoEnabled    = true; 
var sslPrivateKey    = "/etc/pki/tls/private/appx.key";
var sslCertificate   = "/etc/pki/tls/certs/appx.com.crt";
var sslCertAuthority = "/etc/pki/tls/certs/gd_bundle.crt";

var appxdebug = false;
var appxlog = false;

//  *********************************************************
//  Configuration Section - End
//  *********************************************************
var path = require("path");

// If there is an extra argument on the command line that is a request to change our working directory
if (process.argv.length > 2)
    process.chdir(process.argv[2]);
else if( process.env.APPX_CONNECTOR_DIR )
    process.chdir(process.env.APPX_CONNECTOR_DIR);
    
// Find out which result Buffer.toJSON() returns.  This changed between Node 11.0 and 11.1
var appxToJsonTest = new Buffer("test");
var appxIsToJsonAnArray = appxToJsonTest.toJSON(appxToJsonTest) instanceof Array;

var fs = require('fs');

// create a log file
if (appxlog) {
    var d = new Date();
    var logfile = d.getMonth() + 1 + '' + d.getDate() + '' + d.getFullYear() + '' + d.getHours() + '' + d.getMinutes() + '' + d.getSeconds() + '' + '.log';
    fs.appendFile(logfile, 'creating the logfile', function (err) {
        if (err) {
            console.log("Log File Error:  " + err);
        }
        console.log('Created log file:  ' + logfile);
    });
}

var WebSocket = require('ws').WebSocket;
var WebSocketServer = require('ws').Server;
var net = require('net');
var hexy = require('hexy');
var StringDecoder = require('string_decoder').StringDecoder;
var decoder = new StringDecoder('utf8');
var crypto = require('crypto');
var node_cryptojs = require('node-cryptojs-aes');
var CryptoJS = node_cryptojs.CryptoJS;
var JsonFormatter = node_cryptojs.JsonFormatter;
if (sslEnabled)
    var https = require('https');
else
    var https = require('http');

// Configure the websocket
if (sslEnabled) {
    var options = {
        key: fs.readFileSync(sslPrivateKey),
        cert: fs.readFileSync(sslCertificate),
        ca: fs.readFileSync(sslCertAuthority)
    };

    var app = https.createServer(options, function (req, res) {
    }).listen(connectorPort);
}
else {
    var app = https.createServer(function (req, res) {
    }).listen(connectorPort);
}

// start a websocket server
var wss = new WebSocketServer({
    server: app
});

// globals
var clientnumber = 0;

// Websocket Connection Handler
wss.on('connection', function (ws) {
    ws.send2 = function (s) {
        var m = s;
        if (cryptoEnabled)
            m = CryptoJS.AES.encrypt(s, "APPX", {
                format: JsonFormatter
            }).toString();
        return (this.send(m));
    };

    // Add connection meta data to this message for easier debugging
    logactivity("user connected...");

    //increment the client number to have a way to kill it's session's server loop
    var myid = clientnumber++;

    // Start a server loop specific to the client
    // The scope of variables in Javascript and Node allows this websocket connection ("ws") to be
    // mappped to the APPX socket ("client") created below
    var appxprocessor = new APPXProcessor(ws, myid);

    // Create a socket client to APPX
    var client_appx_socket = new net.Socket();

    // Map a function to cleanup on websocket close event
    ws.on('close', function () {
        logactivity('ws client disconnected... #:' + myid);
        client_appx_socket.destroy();
        appxprocessor.end();
    });

    // Map a function to log on websocket error event
    ws.on('error', function (err) {
        logactivity('ws client error occurred' + err);
    });

    // Add a 'close' event handler for the client socket
    client_appx_socket.on('close', function () {
        logactivity('ALERT:  Connection closed');
    });

    // Map a function to handle data from the APPX server
    client_appx_socket.on('data', function (data) {

        logHexDump("APX->CON", data, "received data from the Appx server...");

        // push data from the server onto the clients buffer(an array)
        // Here we could use a NodeJS Buffer object, but by using a Javascript Array, this code will work in browser
        // allowing for easier debugging
	if ( appxIsToJsonAnArray )
	    appxprocessor.rtndata = appxprocessor.rtndata.concat(data.toJSON(), new Array());
	else
	    appxprocessor.rtndata = appxprocessor.rtndata.concat(data.toJSON()['data'], new Array());

        // if we have received enough bytes to satify the next parser request
        // run the parser before receiving any more data.
        logactivity("received data, needbytes=" + appxprocessor.needbytes + ", curbytes=" + appxprocessor.rtndata.length);
        if (appxprocessor.needbytes <= appxprocessor.rtndata.length) {
            var prcd;
            logactivity("calling processor...");
            try {
                prcd = appxprocessor.appxprocessmsg();
                logactivity("returned from called processor");
            }
            catch (ex) {
                logactivity("appxprocessmsg() failed, ex=" + ex);
            }
        }
    });

    // Add an 'end' event handler for the client socket
    client_appx_socket.on('end', function () {
        logactivity('ALERT:  client_appx_socket disconnected');
    });

    // Map a function to log client error event
    client_appx_socket.on('error', function (err) {
        logactivity('ERROR:  client_appx_socket error \n\n' + err);
        ws.send2('ERROR:  client_appx_socket error' + err);
    });

    // Map a function to handle web client message events
    ws.on('message', function (messageCrypt) {
        var message = messageCrypt;
        if (cryptoEnabled) {
            message = CryptoJS.AES.decrypt(messageCrypt, "APPX", {
                format: JsonFormatter
            });
            message = CryptoJS.enc.Utf8.stringify(message);
        }

        // stub up a message object
        var ms = {};
        var g;

        try {
            ms = JSON.parse(message);
            g = true;
            logactivity("received good message:  " + message);
        }
        catch (appxerror) {
            logactivity("Message:  " + message);
            logactivity(appxerror);
            g = false;
        }

        try {
            if (g) {
                switch (ms.cmd) {
                    case "openfile":
                        // may not use this
                        break;
                    case "appxlogin":
                        //Connect to Appx Connection Manager
                        client_appx_socket.connect(parseInt(ms.args[1]), ms.args[0], function () {
                            logactivity("CONNECTED TO " + ms.args[0] + ":" + ms.args[1]);
                        });
                        //created and send login to APPX
                        var tlogin = new Buffer(331);
                        tlogin.fill();
                        tlogin.write(ms.args[2]);
                        tlogin.write(ms.args[3], 21);
                        client_appx_socket.write(tlogin);
                        break;
                    case "appxreconnect":
                        //Connect to Appx Connection Manager
                        client_appx_socket.connect(parseInt(ms.args[1]), ms.args[0], function () {
                            logactivity("CONNECTED TO " + ms.args[0] + ":" + ms.args[1]);
                        });
                        //created and send login to APPX
                        var treconnect = new Buffer(331);
                        treconnect.fill();
                        treconnect.write(ms.args[2]);
                        treconnect.write(ms.args[3], 21);
                        for (var rc = 0; rc < ms.args[4].length; rc++) {
                            treconnect[320 + rc] = ms.args[4].charCodeAt(rc);
                        }
                        client_appx_socket.write(treconnect);
                        break;
                    case "appxnewsession":
                        //Connect to Appx Connection Manager
                        client_appx_socket.connect(parseInt(ms.args[1]), ms.args[0], function () {
                            logactivity("CONNECTED TO " + ms.args[0] + ":" + ms.args[1]);
                        });
                        //created and send login to APPX
                        var tlogin = new Buffer(331);
                        tlogin.fill();
                        tlogin.write(ms.args[2]);
                        tlogin.write(ms.args[3], 21);
                        //add 32 bytes for host and remap
                        if (ms.args.length > 4) {
                            if (ms.args[4]) {
                                tlogin.write(ms.args[4], 74);
                            }
                            if (ms.args[5]) {
                                tlogin.write(ms.args[5], 77);
                            }
                            if (ms.args[6]) {
                                tlogin.write(ms.args[6], 80);
                            }
                            if (ms.args[7]) {
                                tlogin.write(ms.args[7], 90);
                            }
                        }
                        client_appx_socket.write(tlogin);
                        break;
                    case "appxmessage":
                        var b = new Buffer(ms.args);
                        logactivity("ARGS:  " + b);
                        logHexDump("CLI->APX", b, "Client message to Server...");
                        client_appx_socket.write(b);
                        break;
                    case "appxfileuploadmessage":
                        var b = new Buffer(ms.args);
                        logactivity("ARGS:  " + b);
                        logHexDump("CLI->APX", b, "Client message to Server...");
                        client_appx_socket.write(b);
                        var msrtn = new Message();
                        msrtn.data = ms.data;
                        msrtn.datatype = "object";
                        msrtn.hasdata = "yes";
                        msrtn.haserrors = "no";
                        msrtn.type = "APPXFILEUPLOADPROGRESS";
                        ws.send2(JSON.stringify(msrtn));
                        break;
                    case "appxclipboard":
                        var msg = new Buffer(ms.args[0].length + 6);
                        msg.writeUInt32BE(ms.args[0].length, 0); // group
                        msg.writeUInt8(3, 4); // filler
                        msg.writeUInt8(1, 5); // ?senddata, 0 for now, 1 if data length is > amount server asked for
                        //the actual clipboard data
                        msg.write(ms.args[0], 6);
                        logactivity("ARGS:  " + ms.args + ", MSG: " + msg);
                        logHexDump("CLI->APX", msg, "Client message to Server...");
                        client_appx_socket.write(msg);
                        break;
                    case "appxtoken":
                        var msg = new Buffer(12);
                        appxprocessor.curtoken = new TokenStructure();
                        msg.writeUInt32BE(new DataView(new Uint8Array(ms.args.slice(0, 2)).buffer).getUint16(0), 0); // group
                        msg.writeUInt32BE(0, 4); // st pos
                        msg.writeUInt32BE(0, 8); // max pos
                        logactivity("ARGS: " + ms.args + ", MSG: " + msg);
                        logHexDump("CLI->APX", msg, "Client message to Server...");
                        var header = new Buffer(8);
                        header.writeUInt32BE(3, 0);
                        header.writeUInt8(79, 4);
                        header.fill(0, 5);
                        appxprocessor.curtokencount++;
                        client_appx_socket.write(header);
                        client_appx_socket.write(msg);
                        break;
                    case "appxresource":
                        var b = new Buffer(ms.args);
                        appxprocessor.curresourceOut = new ResourceStructure();
                        var msg = new Buffer(24);
                        var argsArray = ms.args.substring(1).split(",");
                        var arg0Array = argsArray[0].split(".");
                        var arg2Array = argsArray[2].split(".");
                        msg.write(arg0Array[0].slice(0, 3), 0); // ap
                        msg.write(arg0Array[1].slice(0, 2), 3); // ver
                        msg.write(arg0Array[2].slice(0, 8), 5); // cacheid
                        msg.writeUInt8(parseInt("0x" + arg0Array[3].slice(0, 2)), 13); // state
                        msg.writeUInt16BE(parseInt("0x" + arg2Array[1].slice(0, 4)), 14); // id
                        msg.writeUInt32BE(parseInt("0x" + arg2Array[0].slice(0, 8)), 16); // ctx
                        msg.writeUInt8(parseInt("0x" + arg2Array[2].slice(0, 2)), 20); // type
                        msg.fill(0, 21); // filler

                        var header = new Buffer(8);
                        header.writeUInt32BE(12, 0);
                        header.writeUInt8(83, 4);
                        header.fill(0, 5);
                        appxprocessor.curresourcecountOut++;
                        client_appx_socket.write(header);
                        client_appx_socket.write(msg);
                        break;
                    case "appxdate":
                        var msg = new Buffer(18);
                        appxprocessor.cursetfield = new SetFieldStructure();
                        msg.writeUInt8(ms.args[0], 0); // row
                        msg.writeUInt8(ms.args[1], 1); // col
                        msg.write(ms.args[2], 2); // date alpha16 string

                        var header = new Buffer(8);
                        header.writeUInt32BE(18, 0);
                        header.writeUInt8(84, 4);
                        header.fill(0, 5);
                        appxprocessor.cursetfieldcount++;
                        client_appx_socket.write(header);
                        client_appx_socket.write(msg);
                        break;
                        //TOD) - FIXME - REENGINEER show messages
                    case "appxsendshow":
                        var msg = new Buffer(18);
                        appxprocessor.cursetfield = new SetFieldStructure();
                        msg.writeUInt8(ms.args[0], 0); // row
                        msg.writeUInt8(ms.args[1], 1); // col
                        msg.write(ms.args[2], 2); // date alpha16 string

                        var header = new Buffer(8);
                        header.writeUInt32BE(18, 0);
                        header.writeUInt8(84, 4);
                        header.fill(0, 5);
                        appxprocessor.cursetfieldcount++;
                        client_appx_socket.write(header);
                        client_appx_socket.write(msg);
                        break;
		    case "updatelocal":
			fs.readFile("localConnectorUpdate.zip", function (err, data) {
                                if (!err) {
                                    ms.data = data;
                                    ms.hasdata = "yes";
                                    ms.haserrors = "no";
                                    ms.type = "UPDATE";
                                    ws.send2(JSON.stringify(ms));
                                }
                                else {
                                    console.log(err);
                                    ms.hasdata = "no";
                                    ms.haserror = "yes";
                                    ms.error = err;
                                    ws.send2(JSON.stringify(ms));
                                }
                            });			
			break;
                    case "appxinit":
                        // This appxinit message provides the opportunity to push down any assets needed on the client
                        // Send anything needed to start the session
                        var sendfile = function (fi) {
                            fs.readFile(fi, 'utf8', function (err, data) {
                                if (!err) {
                                    ms.data = data;
                                    ms.hasdata = "yes";
                                    ms.haserrors = "no";
                                    ms.type = (fi.toLowerCase().indexOf(".js") != -1 ? "SCRIPT" : "STYLE");
                                    ws.send2(JSON.stringify(ms));
                                }
                                else {
                                    console.log(err);
                                    ms.hasdata = "no";
                                    ms.haserror = "yes";
                                    ms.error = err;
                                    ws.send2(JSON.stringify(ms));
                                }
                            });
                        };

                        // Send default CSS stylesheet
                        //sendfile('APPX.css'); // Do we need to send this?  We have it in the public/stylesheets directory. (pete/2014-03-04)
                        // Send client item handler library code
                        sendfile('appx-client-item.js');
                        // Send client key handler library code
                        sendfile('appx-client-keys.js');
                        // Send client localos handler library code
                        sendfile('appx-client-localos.js');
                        // Send main Javascript Library code to handle APPX server messages like Widgets, Shows, Screens
                        sendfile('appx-client-main.js');
                        // Send client menu handler library code
                        sendfile('appx-client-menu.js');
                        // Send client menu handler library code
                        //sendfile('appx-client-menu-test.js');
                        // Send client resource handler library code
                        sendfile('appx-client-resource.js');
                        // Send client screen handler library code
                        sendfile('appx-client-screen.js');
                        // Send client session handler library code
                        sendfile('appx-client-session.js');
                        // Send client token handler library code
                        sendfile('appx-client-token.js');
                        // Send client token handler library code
                        sendfile('appx-client-setfield.js');
                        // Send client utility library code
                        sendfile('appx-client-util.js');
                        // Send client widget handler library code
                        sendfile('appx-client-widget.js');
                        // Send client options handler library code
                        sendfile('appx-client-options.js');
                        //sendfile('widget.js');//Widget Lib - don't use for now
                        break;
                    default:
                        logactivity('no message');
                }
            }
        }
        catch (appxerror2) {
            logactivity(appxerror2);
        }
    });
});

logactivity("Server Running...");

// Logging function, added tracelevel for verbose logging
// May need to make sure file is writable before logging if too much logging is happening
function logactivity(data) {
    try {
        var delimiter = "\r\n";
        if (appxdebug) {
            console.log(data);
        }
        if (appxlog) {
            fs.appendFile(logfile, delimiter + data, function (err) {
                if (err) {
                    console.log("Log File Error:  " + err);
                }
            });
        }
    }
    catch (e) {
        console.log(e);
    }
}

function consoleLogHexDump(tag, data, desc) {
    var format = {
        format: "twos",
        prefix: "[" + tag + "] "
    }
    console.log("");
    console.log("[" + tag + "] " + desc);
    console.log("[" + tag + "] " + data.length + " bytes logged");
    console.log(hexy.hexy(data, format));
}

function logHexDump(tag, data, desc) {
    if (!appxdebug)
        return;
    var format = {
        format: "twos",
        prefix: "[" + tag + "] "
    }
    logactivity("");
    logactivity("[" + tag + "] " + desc);
    logactivity("[" + tag + "] " + data.length + " bytes logged");
    logactivity(hexy.hexy(data, format));
}

// APPXProcessor. Object
// Handle all server Messages
function APPXProcessor(ws, id) {
        // Create a reference to this instance of the APPXProcessor.
        var self = this;
        self.clientid = id;
        self.ws = ws;
        self.rtndata = [];
        self.loggedin = false;
        self.appxserverversion = null;
        self.init_optimization_flag = null;
        self.override = false;
        self.current_show = null;
        self.widgetCount = 0;
        self.widgets = []; //
        self.items = []; //
        self.rowheightpx = 21;
        self.colwidthpx = 8;
        self.screenrows = 31;
        self.screencols = 128;
        self.needbytes = 4;
        self.curhandler = -1;
        self.curhandlerstep = 0;
        self.serverFeatureMask = 0;
        self.serverExtendedFeatureMask = 0;
        self.curresourceIn = null;
        self.curresourcecountIn = 0;
        self.curresourceOut = null;
        self.curresourcecountOut = 0;
        self.cursetfield = null;
        self.cursetfieldcount = 0;
        self.curtokencount = 0;
        self.menu = null;
        self.cursetclip = null;
        self.current_msg = {
            "header": []
        };

        // Ends a web client's server loop
        this.end = function () {
        };

        // APPX message handlers to send JSON packets to client rather than RAW socket data
        // Server Message Handler Functions
        // Processes server messages
        this.appxprocessmsg = function () {
            while ((this.rtndata.length > 0 && this.rtndata.length >= self.needbytes) || (self.needbytes == 0 && this.curhandler > 0)) {
                switch (this.curhandler) {
                    case -1:
                        { // Login needs to be done
                            this.appxloginhandler();
                            break;
                        }
                    case 0:
                        { // Message Block
                            this.appxheaderhandler();
                            break;
                        }
                    case 6:
                        { // Message Block
                            this.appxreconnecthandler();
                            break;
                        }
                    case 10:
                        { // INIT
                            this.appxinithandler();
                            break;
                        }
                    case 12:
                        { // SHOW
                            this.appxshowhandler();
                            break;
                        }
                    case 14:
                        { // FINISHED
                            this.appxfinishhandler();
                            break;
                        }
                    case 17:
                        { // ATTACH
                            this.appxattachhandler();
                            break;
                        }
                    case 19:
                        { // KEYMAP
                            this.appxkeymaphandler();
                            break;
                        }
                    case 21:
                        { // UPDATE SCREEN
                            this.appxscreenhandler();
                            break;
                        }
                    case 22:
                        { // UPDATE ATTRIBUTE
                            this.appxattributeshandler();
                            break;
                        }
                    case 25:
                        { // PING
                            this.appxpinghandler();
                            break;
                        }
                    case 26:
                        { // LOAD URL
                            this.appxloadurlhandler();
                            break;
                        }
                    case 27:
                        { // UPDATE EXTENDEDATTRIBUTE
                            this.appxextraattributeshandler();
                            break;
                        }
                    case 64:
                        { // CREATE OBJECT
                            //this.appxobjectshandler();
                            this.appxobjectcreatehandler();
                            break;
                        }
                    case 65:
                        { // INVOKE METHOD
                            //this.appxobjectshandler();
                            this.appxobjectinvokehandler();
                            break;
                        }
                    case 66:
                        { // DESTROY OBJECT
                            //this.appxobjectshandler(true);
                            this.appxobjectdestroyhandler();
                            break;
                        }
                    case 68:
                        { // FEATURE EXCHANGE
                            this.appxfeatureshandler();
                            break;
                        }
                    case 69:
                        { // SERVER PROCESS ID
                            this.appxpidhandler();
                            break;
                        }
                    case 71:
                        { // SEND FILE
                            this.appxsendfilehandler();
                            break;
                        }
                    case 72: //TAddField.type
                        { // ITEMS
                            this.appxitemshandler();
                            break;
                        }
                    case 73: //TAddField.type
                        { // ITEMS
                            this.appxreceivefilehandler();
                            break;
                        }
                    case 75:
                        { // WIDGETS
                            this.appxwidgetshandler();
                            break;
                        }
                    case 79:
                        { // TOKEN
                            this.appxtokenhandler();
                            break;
                        }
                    case 81:
                        { // MENUS
                            this.appxmenuhandler();
                            break;
                        }
                    case 83:
                        { // RESOURCE
                            this.appxresourcehandler();
                            break;
                        }
                    case 85:
                        { // SETFIELD
                            this.appxsetfieldhandler();
                            break;
                        }
                    case 87:
                        { // SET CLIPBOARD
                            this.appxsetclipboardhandler();
                            break;
                        }
                    case 88:
                        { // GET CLIPBOARD
                            this.appxgetclipboardhandler();
                            break;
                        }
                    case 89:
                        { // CONSTANTS EXCHANGE
                            this.appxconstantshandler();
                            break;
                        }
                    case 91:
                        { // GET MESSAGES
                            this.appxgetmessageshandler();
                            break;
                        }
                    case 94:
                        { // EXTENDED FEATURE EXCHANGE
                            this.appxextendedfeatureshandler();
                            break;
                        }
                    default:
                        {
                            //exit;
                            break;
                        }
                }
            }
            logactivity("appxprocessmsg exiting, needbytes=" + this.needbytes + "gotbytes=" + this.rtndata.length);
        };

        //<20140523_chris@praclox.nl_javascan>
        /* TODO: client side scan list (single column)
        sub 0la NSCAN JAVA
        :CREATE CLIENT SCAN OBJECT
            ,RT_CLI_CREATE_OBJECT:  view.AppxTokenScan()
            TCreateObject.read: MObjBlock, MArgBlock[]
            doc.createLocalObject(TCreateObject)
            TCreateObject.write: MCreateReturn(int handle, int error)
        :ADD TO CLIENT SCAN LIST
            ,RT_CLI_INVOKE_METHOD:  void addListItem(String)
            TInvokeMethod.read: MObjBlock, MArgBlock[]
            doc.invokeLocalMethod(TInvokeMethod)
            TInvokeMethod.write: MInvokeReturn(int error);
        :SHOW THE CLIENT SCAN LIST
            ,RT_CLI_INVOKE_METHOD:  void showTheList()
            TInvokeMethod.read: MObjBlock, MArgBlock[]
            doc.invokeLocalMethod(TInvokeMethod)
            TInvokeMethod.write: MInvokeReturn(int error);
        :WAIT FOR OBJECT EVENT
            AppxTokenScan.mouseClicked
                IOHandler.fireObjectEvent(int handle, int return_code)
                    TObjectEvent
        :DESTROY THE CLIENT SCAN LIST
            ,RT_CLI_DESTROY_OBJECT
            TDestroyObject.read: MObjBlock
        */
        var T_BOOLEAN = 1; //1
        var T_BYTE = 2; //1
        var T_CHAR = 3; //1
        var T_DOUBLE = 4; //8
        var T_FLOAT = 5; //4
        var T_INT = 6; //4
        var T_LONG = 7; //4
        var T_SHORT = 8; //2
        var T_UNSIGNED_BYTE = 9; //1
        var T_UNSIGNED_SHORT = 10; //2
        var T_STRING = 11;
        var T_DATE = 12; //new SimpleDateFormat("yyyyMMddHHmmssSS").parse

        this.dataType = null;
        this.data = function () {
            var ret = this.rtndata.slice(0, this.needbytes); //STRING,DATE
            if (this.dataType && ret.length > 0) {
                var dv = new DataView(new Uint8Array(ret).buffer);
                switch (this.dataType) {
                    case T_BYTE:
                        ret = dv.getInt8(0);
                        break;
                    case T_UNSIGNED_BYTE:
                    case T_BOOLEAN:
                    case T_CHAR:
                        ret = dv.getUint8(0);
                        break;
                    case T_SHORT:
                        ret = dv.getInt16(0);
                        break;
                    case T_UNSIGNED_SHORT:
                        ret = dv.getUint16(0);
                        break;
                    case T_INT:
                        ret = dv.getInt32(0); //UINT:getUint32
                        break;
                    case T_FLOAT:
                        ret = dv.getFloat32(0);
                        break;
                    case T_DOUBLE:
                    case T_LONG: //ULONG
                        ret = dv.getFloat64(0);
                        break;
                    default:
                        ret = ab2str(ret).trim();
                }
                this.dataType = null;
            }
            this.rtndata = this.rtndata.slice(this.needbytes); //,this.rtndata.length
            return ret;
        };
        this.done = function () {
            this.curhandler = 0;
            this.curhandlerstep = 0;
            this.needbytes = 8;
        };

        this.read = function (len, type) {
            this.dataType = (type ? type : T_STRING);
            this.curhandlerstep++;
            this.needbytes = len;
        };
        this.readByte = function () {
            this.read(1, T_BYTE);
        };
        this.readShort = function () {
            this.read(2, T_SHORT);
        };
        this.readInt = function () {
            this.read(4, T_INT);
        };

        this.send = function (type, data) {
            var ms = new Message();
            ms.type = type;
            if (data) {
                ms.data = data;
                ms.datatype = "object";
                ms.hasdata = true;
            }
            ms.haserrors = false;
            ms.errormsg = "";
            ws.send2(JSON.stringify(ms));
            this.done();
        };

        this.step = function () {
            return this.curhandlerstep;
        };
        this.stepTo = function (step) {
            this.curhandlerstep = step;
            this.needbytes = 0;
        };

        this.hdr = null;
        this.arg = null;
        this.appxobjecthandler = function (type) {
            switch (this.step()) {
                case 0:
                    this.hdr = {}; //new MObjBlock();
                    this.hdr.argBlocks = [];
                    this.readShort();
                    break;
                case 1:
                    this.hdr.protoLen = this.data();
                    this.readByte();
                    break;
                case 2:
                    this.hdr.argCount = this.data();
                    this.readByte();
                    break;
                case 3:
                    this.hdr.handle = this.data();
                    this.read(this.hdr.protoLen);
                    break;
                case 4:
                    this.hdr.proto = this.data();
                    if (this.hdr.argCount == 0) this.stepTo(9);
                    else this.read(0);
                    break;
                case 5:
                    this.arg = {}; //new MArgBlock();
                    this.readInt();
                    break;
                case 6:
                    //tmnetsrv actually only supports ELEM_ALP_CONTIG (string) atm
                    this.arg.dataType = this.data();
                    this.readInt();
                    break;
                case 7:
                    this.arg.dataLength = this.data();
                    this.read(this.arg.dataLength, this.arg.dataType);
                    break;
                case 8:
                    this.arg.dataObject = this.data();
                    this.hdr.argBlocks.push(this.arg);
                    if (--this.hdr.argCount > 0) this.stepTo(5);
                    else this.read(0); //stepTo(9)
                    break;
                case 9:
                    this.send("APPX" + type + "OBJECT", this.hdr);
                    break;
            }
        };
        this.appxobjectcreatehandler = function () {
            this.appxobjecthandler("CREATE");
        }; //64
        this.appxobjectinvokehandler = function () {
            this.appxobjecthandler("INVOKE");
        }; //65
        this.appxobjectdestroyhandler = function () {
            this.appxobjecthandler("DESTROY");
        }; //66
        //</20140523_chris@praclox.nl_javascan>

        // Objects Handler
        /*this.appxobjectshandler = function () {

            logactivity("***** CREATE_OBJECT *****, step=" + this.curhandlerstep);
            switch (this.curhandlerstep) {
                case 0: // set up to read an item block from the server
                    this.currentjavaobject = new JavaObject();

                    //get structure bytes
                    this.currentjavaobject.createobjectstructurebytes = this.rtndata.slice(0, this.needbytes);
                    this.currentjavaobject.createobjectstructurebytes = this.rtndata.slice(0, 4);

                    //remove used bytes from rtndata
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);

                    //set available variables from createobjectstructurebytes
                    this.currentjavaobject.prototypelength = new DataView(new Uint8Array(this.currentjavaobject.createobjectstructurebytes.slice(0, 2)).buffer).getUint16(0);
                    this.currentjavaobject.argumentcount = new DataView(new Uint8Array(this.currentjavaobject.createobjectstructurebytes.slice(2, 3)).buffer).getUint8(0);
                    this.currentjavaobject.objecthandle = new DataView(new Uint8Array(this.currentjavaobject.createobjectstructurebytes.slice(3, 4)).buffer).getUint8(0);

                    //set up needbytes for next step
                    this.needbytes = this.currentjavaobject.prototypelength;

                    this.curhandlerstep = 1;
                    break;
                case 1:

                    //get prototype data from buffer
                    this.currentjavaobject.prototypedata = this.rtndata.slice(0, this.needbytes);

                    //remove used bytes from buffer
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    if (this.currentjavaobject.argumentcount > 0) {

                        //set needbytes to the first argument structure length(always 8)
                        this.needbytes = 8;
                        this.curhandlerstep = 2;

                    } else {

                        //no arguments, send object to client

                        this.curhandler = 0;
                        this.needbytes = 8;
                        this.curhandlerstep = 0;

                        var ms = new Message();
                        ms.type = "APPXCREATEOBJECT";
                        ms.data = this.currentjavaobject;
                        ms.datatype = "object";
                        ms.hasdata = true;
                        ms.haserrors = false;
                        ms.errormsg = "";
                        ws.send2(JSON.stringify(ms));

                    }

                    break;
                case 2:

                    //self.argumentstructurebytes = null;
                    //self.argumentdatatype = null;
                    //self.argumentdatalength = null;

                    this.currentjavaobject.argumentstructurebytes = this.rtndata.slice(0, this.needbytes);

                    //remove used bytes from buffer
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    this.currentjavaobject.argumentdatatype = new DataView(new Uint8Array(this.currentjavaobject.argumentstructurebytes.slice(0, 4).buffer)).getUint32(0);
                    this.currentjavaobject.argumentdatalength = new DataView(new Uint8Array(this.currentjavaobject.argumentstructurebytes.slice(4, 8).buffer)).getUint32(0);

                    this.needbytes = this.currentjavaobject.argumentdatalength;

                    //
                    this.curhandlerstep = 3;

                    break;
                case 3:

                    //self.argumentdata = null;
                    this.currentjavaobject.argumentdata = this.rtndata.slice(0, this.needbytes);

                    //remove used bytes from buffer
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    //now we have the complete argument,
                    //add it to the arguments array on the currentjavaobject
                    this.currentjavaobject.arguments.push({ "datatype": this.currentjavaobject.argumentdatatype, "length": this.currentjavaobject.argumentdatalength, "data": this.currentjavaobject.argumentdata });

                    if (this.currentjavaobject.arguments.length < this.currentjavaobject.argumentcount) {

                        //more args coming, set needbytes to length of arg structure
                        this.needbytes = 8;

                        //send loop back to step 2 to get next argument
                        this.curhandlerstep = 2;

                    } else {

                        //no more arguments, send object to client

                        //reset and send msg to client
                        this.curhandler = 0;
                        this.needbytes = 8;
                        this.curhandlerstep = 0;

                        var ms = new Message();
                        ms.type = "APPXCREATEOBJECT";
                        ms.data = this.currentjavaobject;
                        ms.datatype = "object";
                        ms.hasdata = true;
                        ms.haserrors = false;
                        ms.errormsg = "";
                        ws.send2(JSON.stringify(ms));

                    }

                    break;
            }
        };*/

        // Login Handler
        this.appxloginhandler = function () {
            logactivity("***** SOCKET_LOGIN *****");

            switch (this.curhandlerstep) {
                case 0:
                    var loginresponse = this.rtndata.slice(0, 2);
                    this.rtndata = this.rtndata.slice(2, this.rtndata.length);

                    if (loginresponse[1] == 1) {

                        logactivity("logged in...");
                        this.loggedin = true;
                        this.appxserverversion = this.rtndata.slice(0, 2);
                        this.rtndata = this.rtndata.slice(2, this.rtndata.length);

                        var ms = new Message();
                        ms.type = "APPXLOGIN";
                        ms.data = "The user logged in successfully!";
                        ms.datatype = "string";
                        ms.hasdata = true;
                        ms.transresult = "SUCCESS";
                        ms.haserrors = false;
                        ms.errormsg = "";
                        ws.send2(JSON.stringify(ms));

                        this.curhandler = 0;
                        this.curhandlerstep = 0;
                        this.needbytes = 8;
                    }
                    else {
                        this.curhandlerstep = 1;
                        this.needbytes = 4;
                    }

                    break;
                case 1:

                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getInt32(0);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);
                    this.curhandlerstep = 2;

                    break;
                case 2:

                    var msg = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    var ms = new Message();
                    ms.type = "APPXLOGIN";
                    ms.data = ab2str(msg);
                    ms.datatype = "string";
                    ms.hasdata = true;
                    ms.transresult = "FAIL";
                    ms.haserrors = true;
                    ms.errormsg = "Login failed";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.curhandlerstep = 0;
                    this.needbytes = 0;

                    break;
                default:
                    break;
            }

        };

        // Reconnect Handler
        this.appxreconnecthandler = function () {
            logactivity("***** SOCKET_RECONNECT *****");
            switch (this.curhandlerstep) {
                case 0:
                    var bytes = this.rtndata.slice(0, 4);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);
                    this.recondatalength = new DataView(new Uint8Array(bytes).buffer).getUint32(0);
                    this.curhandlerstep = 1;
                    this.needbytes = this.recondatalength;
                    break;
                case 1:
                    var bytes = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    //send the row cols to the client

                    var ms = new Message();
                    ms.data = bytes;
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXRECONNECT";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.curhandlerstep = 0;
                    this.needbytes = 8;
                    break;
            }

        };

        this.appxfinishhandler = function () {
            this.loggedin = false;
            this.curhandler = -1;
            this.curhandlerstep = 0;
            this.needbytes = 0;
            this.rtndata = this.rtndata.slice();
            var ms = new Message();
            ms.type = "APPXFINISH";
            ws.send2(JSON.stringify(ms));
        };

        // Header Handler
        this.appxheaderhandler = function () {
            logactivity("***** SOCKET_HEADER *****");

            self.current_msg.header = this.rtndata.slice(0, 8);
            this.rtndata = this.rtndata.slice(8, this.rtndata.length);
            var constants_bytes = self.current_msg.header.slice(0, 4);
            var constants_length = new DataView(new Uint8Array(constants_bytes).buffer).getUint32(0);

            self.needbytes = constants_length;
            self.curhandler = self.current_msg.header[4];
            self.curhandlerstep = 0;

            logactivity("header processed, length=" + self.needbytes + ", command=" + self.curhandler);
        };

        // Init Handler
        this.appxinithandler = function (data) {
            logactivity("***** SOCKET_INIT *****, current step=" + this.curhandlerstep);
            switch (this.curhandlerstep) {
                case 0:
                    this.needbytes = 1;
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    this.init_optimization_flag = this.rtndata.slice(0, 1);
                    this.rtndata = this.rtndata.slice(1, this.rtndata.length);

                    var ms = new Message();
                    ms.data = data;
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXINIT";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.needbytes = 8;
                    this.curhandlerstep = 0;
                    break;
            }
        };

        // Features Handler
        this.appxfeatureshandler = function (data) {
            logactivity("***** SOCKET_FEATURE_EXCHANGE *****");
            var bytes = this.rtndata.slice(0, 4);
            this.rtndata = this.rtndata.slice(4, this.rtndata.length);

            this.serverFeatureMask = new DataView(new Uint8Array(bytes).buffer).getUint32(0);

            var ms = new Message();
            ms.data = data;
            ms.hasdata = "yes";
            ms.haserrors = "no";
            ms.type = "APPXFEATURES";
            ws.send2(JSON.stringify(ms));

            logactivity("server feature mask=" + this.serverFeatureMask);

            this.curhandler = 0;
            this.needbytes = 8;
            this.curhandlerstep = 0;
        };

        // Extended Features Handler //94
        this.appxextendedfeatureshandler = function (data) {
            logactivity("***** SOCKET_EXTENDED_FEATURE_EXCHANGE *****");
            var bytes = this.rtndata.slice(0, 4);
            this.rtndata = this.rtndata.slice(4, this.rtndata.length);

            this.serverExtendedFeatureMask = new DataView(new Uint8Array(bytes).buffer).getUint32(0);

            var ms = new Message();

            ms.data = data;
            ms.hasdata = "yes";
            ms.haserrors = "no";
            ms.type = "APPXEXTENDEDFEATURES";

            ws.send2(JSON.stringify(ms));

            logactivity("server extended feature mask=" + this.serverExtendedFeatureMask);

            this.curhandler = 0;
            this.needbytes = 8;
            this.curhandlerstep = 0;
        };

        // Constants Handler
	/*
        this.appxconstantshandler = function (data) {
            var ms = new Message();
            ms.data = data;
            ms.hasdata = "yes";
            ms.haserrors = "no";
            ms.type = "APPXCONSTANTS";
            ws.send2(JSON.stringify(ms));
        };
	*/

        // Attach Handler
        this.appxattachhandler = function (data) {
            logactivity("***** SOCKET_ATTACH *****");
            var ms = new Message();
            ms.data = data;
            ms.hasdata = "yes";
            ms.haserrors = "no";
            ms.type = "APPXATTACH";
            ws.send2(JSON.stringify(ms));

            this.curhandler = 0;
            this.needbytes = 8;
            this.curhandlerstep = 0;
        };

        // KeyMap Handler
        this.appxkeymaphandler = function (data) {
            logactivity("***** SOCKET_KEYMAP *****, step=" + this.curhandlerstep);

            switch (this.curhandlerstep) {
                case 0:
                    this.needbytes = 10;
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    this.mykeymapkey = this.rtndata.slice(0, 10);
                    this.rtndata = this.rtndata.slice(10, this.rtndata.length);
                    this.needbytes = 2;
                    this.curhandlerstep = 2;
                    break;
                case 2:
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 2)).buffer).getUint16(0);
                    this.rtndata = this.rtndata.slice(2, this.rtndata.length);
                    this.curhandlerstep = 3;
                    break;
                case 3:
                    if (this.needbytes > 0) {
                        self.mykeymapdata = rtndata.slice(0, this.needbytes);
                        this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    }

                    var ms = new Message();
                    ms.data = data;
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXKEYMAP";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.needbytes = 8;
                    this.curhandlerstep = 0;
                    break;
            }
        };

        // PID Handler
        this.appxpidhandler = function () {
            logactivity("***** SOCKET_SERVER_PID *****, step=" + this.curhandlerstep);

            switch (this.curhandlerstep) {
                case 0:
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 2)).buffer).getUint16(0);
                    this.rtndata = this.rtndata.slice(2, this.rtndata.length);

                    this.curhandlerstep = 1;
                    break;
                case 1:
                    var bdlen = 0;
                    var mypid = 0;
                    var mypidbytes = [];

                    mypidbytes = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    mypid = mypidbytes;

                    logactivity("my process id=" + mypid);

                    var ms = new Message();
                    ms.data = mypid;
                    ms.datatype = "object";
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXPID";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.needbytes = 8;
                    this.curhandlerstep = 0;
                    break;
            }
        };

        // SEND FILE Handler
        this.appxsendfilehandler = function () {
            logactivity("got a file message...");
            switch (this.curhandlerstep) {
                case 0: // set up to read an item block from the server
                    this.currfile = new FileStructure();
                    this.needbytes = 4;
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    this.currfile.datalength = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getUint32(0);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);
                    this.needbytes = 4;
                    this.curhandlerstep = 2;
                    break;
                case 2:
                    this.currfile.filenamelength = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getUint32(0);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);
                    this.needbytes = this.currfile.filenamelength;
                    this.curhandlerstep = 3;
                    break;
                case 3:
                    this.currfile.filename = ab2str(this.rtndata.slice(0, this.needbytes));
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    this.needbytes = 4;
                    this.curhandlerstep = 4;

                    //send the file name on to the client so it can send a status structure
                    //whether or not to send the rest of the file.
                    logactivity("sending filename to client:" + this.currfile.filename);

                    var ms = new Message();
                    ms.data = this.currfile;
                    ms.datatype = "object";
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.messagepart = 3; // 3 = filename
                    ms.type = "APPXSENDFILE";
                    ws.send2(JSON.stringify(ms));
                    // next is data chunk length - 4 bytes
                    break;
                case 4:
                    // may need a timeout here or in the beginning of the handler
                    // in case the client doesn't respond with the status structure
                    this.currfile.currdatachunklength = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getUint32(0);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);

                    if (this.currfile.currdatachunklength > 0) {
                        this.needbytes = this.currfile.currdatachunklength;
                        this.curhandlerstep = 5;
                    }
                    else {
                        // send file EOF
                        var ms = new Message();
                        ms.data = this.currfile.currdatachunklength;
                        ms.datatype = "EOF";
                        ms.hasdata = "yes";
                        ms.haserrors = "no";
                        ms.messagepart = -1;
                        ms.type = "APPXSENDFILE";
                        ws.send2(JSON.stringify(ms));
                        //exit this handler, reset to header handler
                        this.curhandler = 0;
                        this.needbytes = 8;
                        this.curhandlerstep = 0;
                    }

                    break;

                case 5:

                    this.currfile.currdatachunk = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    // always go back to step 4 which handles EOF block
                    this.needbytes = 4;
                    this.curhandlerstep = 4;

                    // send file data chunk
                    // we could store and send all at once??
                    var ms = new Message();
                    ms.data = this.currfile.currdatachunk;
                    ms.datatype = "array";
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.messagepart = 5; // 5 = data chunk
                    ms.type = "APPXSENDFILE";
                    ws.send2(JSON.stringify(ms));

                    break;
            }
        };

        // LOAD URL Handler
        this.appxloadurlhandler = function () {
            logactivity("***** SOCKET_LOAD_URL *****, step=" + this.curhandlerstep);
            switch (this.curhandlerstep) {
                case 0: // set up to read an item block from the server
                    this.needbytes = 2;
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 2)).buffer).getUint16(0);
                    this.rtndata = this.rtndata.slice(2, this.rtndata.length);

                    this.curhandlerstep = 2;
                    break;
                case 2:
                    var bdlen = 0;
                    var myurlbytes = [];

                    myurlbytes = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    logactivity("my url=" + ab2str(myurlbytes));

                    var ms = new Message();
                    ms.data = ab2str(myurlbytes).trim();
                    ms.datatype = "object";
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXLOADURL";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.needbytes = 8;
                    this.curhandlerstep = 0;
                    break;
            }
        };

        // MENUS Handler
        this.appxmenuhandler = function () {

            logactivity("***** SOCKET_RESOURCE *****, step=" + this.curhandlerstep);

            var done = false;
            var menuItemDone = false;

            var nb = this.needbytes;
            switch (this.curhandlerstep) {

                case 0: // set up to read an menu headeer block from the server
                    this.menu = new Menu();
                    this.curhandlerstep++;
                    this.needbytes = 6;
                    break;

                case 1:
                    this.menu.headerstructure = this.rtndata.slice(0, 6);
                    //one byte can use rthdata.shift, but it might get confusing
                    this.menu.type = this.menu.headerstructure.slice(0, 1);
                    this.menu.itemcount = new DataView(new Uint8Array(this.menu.headerstructure.slice(2, 4)).buffer).getUint16(0);
                    this.menu.headerdatalength = new DataView(new Uint8Array(this.menu.headerstructure.slice(4, 6)).buffer).getUint16(0);

                    if (this.menu.headerdatalength > 0) {
                        this.curhandlerstep++;
                        this.needbytes = this.menu.headerdatalength;
                    }
                    else { // if menu datalength is 0 skip a step
                        this.curhandlerstep += 2;
                        this.needbytes = 0;
                    }
                    break;

                case 2: //fetch header
                    this.menu.headerdata = this.rtndata.slice(0, this.needbytes);
                    //this.menu.headerdataparams = this.appxparamsparser(this.menu.headerdata);
                    this.curhandlerstep++;
                    this.needbytes = 0;
                    break;

                case 3: //are there any menu items?
                    if (this.menu.itemcount > 0) {
                        this.curhandlerstep++;
                        this.needbytes = 4; //menu struct = 4 bytes
                    }
                    else {
                        done = true;
                        menuItemDone = true;
                    }
                    break;

                case 4: //get next menu item (MMnu)
                    this.curmenuitem = new MenuItem();
                    this.curmenuitem.structure = this.rtndata.slice(0, 4);
                    this.curhandlerstep++;
                    this.needbytes = new DataView(new Uint8Array(this.curmenuitem.structure.slice(2, 4)).buffer).getUint16(0);
                    break;

                case 5: //process menu item
                    this.curmenuitem.data = this.rtndata.slice(0, this.needbytes);
                    this.curhandlerstep++;
                    this.needbytes = 4;
                    break;

                case 6: // trim off and process extra data length
                    // FIXME: add code here
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getInt32(0);
                    if (this.needbytes > 0) {
                        this.curhandlerstep++;
                    }
                    else {
                        if (this.needbytes < 0) {
                            this.curmenuitem.extrareuse = true;
                        }
                        menuItemDone = true;
                    }
                    break;
                case 7: // trim off and process extra data
                    // FIXME: add code here
                    this.curmenuitem.extradata = this.rtndata.slice(0, this.needbytes);
                    menuItemDone = true;
                    break;
            }
            var hexdata = this.rtndata.slice(0, nb);
            logHexDump("ITEMDATA", hexdata, "item data");
            if (this.curhandlerstep > 1) this.rtndata = this.rtndata.slice(nb);

            if (menuItemDone) {
                this.menu.items.push(this.curmenuitem);
                if (this.menu.items.length < this.menu.itemcount) {
                    this.curhandlerstep = 4;
                    this.needbytes = 4;
                }
                else {
                    var ms = new Message();
                    ms.data = this.menu;
                    ms.datatype = "array";
                    ms.datacount = this.menu.items.length;
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXMENU";
                    ws.send2(JSON.stringify(ms));
                    done = true;
                }
            }

            if (done) {
                // setup to read next message header
                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
                logHexDump("MENUDATA", "curmenu", "Resource Data Received...");
            }
        };

        // TOKEN Handler
        this.appxtokenhandler = function () {

            logactivity("***** SOCKET_TOKEN *****, step=" + this.curhandlerstep);

            var done = false;

            switch (this.curhandlerstep) {
                case 0:
                    this.curtokencount--;
                    var rawdata = this.rtndata.slice(0, 12);
                    this.rtndata = this.rtndata.slice(12, this.rtndata.length);

                    logHexDump("resHeader", rawdata, "Resource header...");
                    this.curtoken.grp = rawdata.slice(0, 4);
                    this.curtoken.start = rawdata.slice(4, 8);
                    this.curtoken.max = rawdata.slice(8, 12);
                    this.curtoken.items = [];

                    this.needbytes = 8;
                    this.curhandlerstep = 1;

                    break;

                case 1:
                    this.curtoken.totalsize = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getUint8(0);
                    this.curtoken.len = new DataView(new Uint8Array(this.rtndata.slice(4, 8)).buffer).getUint32(0);
                    this.rtndata = this.rtndata.slice(8, this.rtndata.length);

                    if (this.curtoken.len > 0) {
                        this.needbytes = 8;
                        this.curhandlerstep = 2;
                    }
                    else {
                        done = true;
                    }

                    break;
                case 2:
                    this.curtoken.data = [];

                    this.curtoken.data = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    var ti = new TokenItem();
                    ti.token_id = new DataView(new Uint8Array(this.curtoken.data.slice(0, 4)).buffer).getUint32(0);
                    ti.token_id_group = new DataView(new Uint8Array(this.curtoken.data.slice(4, 6)).buffer).getUint16(0);
                    ti.token_id_val_len = new DataView(new Uint8Array(this.curtoken.data.slice(6, 8)).buffer).getUint16(0);
                    if (ti.token_id_val_len != 0) {
                        this.needbytes = ti.token_id_val_len;
                        this.curhandlerstep = 3;
                    }
                    else {
                        if (this.curtoken.items.length = this.curtoken.len) {
                            done = true;
                        }
                        else {
                            this.needbytes = 8;
                            this.curhandlerstep = 2;
                        }
                    }
                    this.curtoken.items.push(ti);
                    break;
                case 3:
                    var ti = this.curtoken.items[this.curtoken.items.length - 1];

                    ti.data = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    if (this.curtoken.items.length == this.curtoken.len) {
                        done = true;
                    }
                    else {
                        this.needbytes = 8;
                        this.curhandlerstep = 2;
                    }
                    break;
            }

            if (done) {
                logHexDump("TOKENDATA", this.curtoken.data, "Resource Data Received...");

                var ms = new Message();
                ms.data = this.curtoken;
                ms.datatype = "object";
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXTOKEN";
                ws.send2(JSON.stringify(ms));

                logactivity("curtokencount=" + this.curtokencount);
                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }
        };

        // RESOURCE Handler
        this.appxresourcehandler = function () {
            logactivity("***** SOCKET_RESOURCE *****, step=" + this.curhandlerstep);

            var done = false;

            switch (this.curhandlerstep) {
                case 0:
                    this.needbytes = 20;
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    this.curresourcecountIn--;
                    var rawdata = this.rtndata.slice(0, 20);
                    this.rtndata = this.rtndata.slice(20, this.rtndata.length);

                    logHexDump("resHeader", rawdata, "Resource header...");
                    if (this.curresourceIn == null)
                        this.curresourceIn = new ResourceStructure();
                    this.curresourceIn.ap = rawdata.slice(0, 3);
                    this.curresourceIn.ver = rawdata.slice(3, 5);
                    this.curresourceIn.cacheid = rawdata.slice(5, 13);
                    this.curresourceIn.state = new DataView(new Uint8Array(rawdata.slice(13, 14)).buffer).getUint8(0);
                    this.curresourceIn.loctype = new DataView(new Uint8Array(rawdata.slice(14, 15)).buffer).getUint8(0);
                    this.curresourceIn.ext = new DataView(new Uint8Array(rawdata.slice(15, 16)).buffer).getUint8(0);
                    this.curresourceIn.len = new DataView(new Uint8Array(rawdata.slice(16, 20)).buffer).getUint32(0);
                    this.curresourceIn.data = [];

                    if (this.curresourceIn.len > 0) {
                        this.needbytes = this.curresourceIn.len;
                        this.curhandlerstep = 2;
                    }
                    else {
                        done = true;
                    }

                    break;
                case 2:
                    this.curresourceIn.data = [];
                    this.curresourceIn.data = base64Array(this.rtndata.slice(0, this.needbytes));
                    //this.curresourceIn.data = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    done = true;
                    break;
            }

            if (done) {
                logHexDump("RESDATA", this.curresourceIn.data, "Resource Data Received...");

                var ms = new Message();
                ms.data = this.curresourceIn;
                ms.datatype = "object";
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXRESOURCE";
                ws.send2(JSON.stringify(ms));

                logactivity("curresourcecount=" + this.curresourcecountIn);
                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }
        };

        // SET CLIPBOARD Handler - 87
        this.appxsetclipboardhandler = function () {

            logactivity("***** SOCKET_SETCLIPBOARD *****, step=" + this.curhandlerstep);

            var done = false;

            switch (this.curhandlerstep) {
                case 0:

                    this.cursetclip = {};

                    this.needbytes = 4;
                    this.curhandlerstep = 1;

                    break;

                case 1:

                    //rtndata was empty once, so putting in a check
                    var rawdata = this.rtndata.slice(0, 4);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);

                    this.cursetclip.datatype = new DataView(new Uint8Array(rawdata.slice(0, 4)).buffer).getUint32(0);

                    this.needbytes = 4;
                    this.curhandlerstep = 2;
                    break;


                case 2:

                    //rtndata was empty once, so putting in a check
                    var rawdata = this.rtndata.slice(0, 4);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);

                    this.cursetclip.datalength = new DataView(new Uint8Array(rawdata.slice(0, 4)).buffer).getUint32(0);

                    this.needbytes = this.cursetclip.datalength;
                    this.curhandlerstep = 3;
                    break;

                case 3:


                    this.cursetclip.data = ab2str(this.rtndata.slice(0, this.needbytes));
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    done = true;

                    break;
            }

            if (done) {
                logHexDump("SETCLIPBOARDDATA", this.cursetclip.data, "GetClipboard Data Message Received...");

                var ms = new Message();
                ms.data = this.cursetclip;
                ms.datatype = "string";
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXSETCLIPBOARD";
                ws.send2(JSON.stringify(ms));

                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }
        };

        // GET CLIPBOARD Handler - 88
        this.appxgetclipboardhandler = function () {
            logactivity("***** SOCKET_GETCLIPBOARD *****, step=" + this.curhandlerstep);

            var done = false;

            switch (this.curhandlerstep) {
                case 0:

                    this.needbytes = 9;
                    this.curhandlerstep = 1;

                    break;

                case 1:
                    var rawdata = this.rtndata.slice(0, 9);
                    this.rtndata = this.rtndata.slice(9, this.rtndata.length);

                    this.clipboardstruct = {};
                    this.clipboardstruct.datatype = new DataView(new Uint8Array(rawdata.slice(0, 4)).buffer).getUint32(0);
                    this.clipboardstruct.datalength = new DataView(new Uint8Array(rawdata.slice(4, 8)).buffer).getUint32(0);
                    this.clipboardstruct.senddata = rawdata.slice(8, 1);

                    done = true;
                    break;
            }

            if (done) {
                logHexDump("GETCLIPBOARDDATA", "", "GetClipboard Data Message Received...");

                var ms = new Message();
                ms.data = this.clipboardstruct;
                ms.datatype = "string";
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXGETCLIPBOARD";
                ws.send2(JSON.stringify(ms));

                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }
        };

        // CONSTANTS EXCHANGE Handler - 89
        this.appxconstantshandler = function () {
            logactivity("***** SOCKET_CONSTANTS EXCHANGE *****, step=" + this.curhandlerstep);

            var done = false;
	    
            switch (this.curhandlerstep) {
                case 0: // get the first header
                    this.constants = {};
                    this.needbytes = 4;
                    this.curhandlerstep = 1;

                    break;

                case 1: // process header
                    var rawdata = this.rtndata.slice(0, 4);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);
		    done = true;
                    break;
            }

            if (done) {
                var ms = new Message();
                ms.data = [];
                ms.datatype = "array";
                ms.datacount = 0;
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXCNSTS";
                ws.send2(JSON.stringify(ms));

                // setup to read next message header
                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }
        };

        // GET MESSAGES Handler - 91
        this.appxgetmessageshandler = function () {
            logactivity("***** SOCKET_GETMESSAGES *****, step=" + this.curhandlerstep);

            var done = false;
            var allDone = false;

            switch (this.curhandlerstep) {
                case 0: // get the first header
                    this.messages = [];
                    this.needbytes = 8;
                    this.curhandlerstep = 1;

                    break;

                case 1: // process header
                    var rawdata = this.rtndata.slice(0, 8);
                    this.rtndata = this.rtndata.slice(8, this.rtndata.length);

                    var msg = new Mesg();
                    msg.group = new DataView(new Uint8Array(rawdata.slice(0, 4)).buffer).getUint32(0);
                    msg.severity = new DataView(new Uint8Array(rawdata.slice(4, 6)).buffer).getUint16(0);
                    msg.txtlen = new DataView(new Uint8Array(rawdata.slice(6, 8)).buffer).getUint16(0);
                    this.messages.push(msg);

                    if (msg.txtlen > 0) {
                        this.needbytes = msg.txtlen;
                        this.curhandlerstep = 2;
                    }
                    else {
                        done = true;
                    }
                    break;

                case 2: // process text
                    var msg = this.messages[this.messages.length - 1];
                    var rawdata = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    msg.txtval = ab2str(rawdata);
                    done = true;
            }

            if (done) {
                var msg = this.messages[this.messages.length - 1];
                if (msg.group == 0 && msg.severity == 0 && msg.txtlen == 0) {
                    allDone = true;
                }
                else {
                    this.needbytes = 8;
                    this.curhandlerstep = 1;
                }
            }

            if (allDone) {
                logactivity("processing " + this.items.length + " messages...");

                var ms = new Message();
                ms.data = this.messages;
                ms.datatype = "array";
                ms.datacount = this.messages.length;
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXMSGS";
                ws.send2(JSON.stringify(ms));

                // setup to read next message header
                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }

        };

        // SETFIELD Handler //85
        this.appxsetfieldhandler = function () {

            logactivity("***** SOCKET_SETFIELD *****, step=" + this.curhandlerstep);

            var done = false;

            switch (this.curhandlerstep) {
                case 0:
                    this.cursetfieldcount--;
                    var rawdata = this.rtndata.slice(0, 4);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);


                    logHexDump("resHeader", rawdata, "Resource header...");

                    this.cursetfield.row = new DataView(new Uint8Array(rawdata.slice(0, 1)).buffer).getUint8(0);
                    this.cursetfield.col = new DataView(new Uint8Array(rawdata.slice(1, 2)).buffer).getUint8(0);
                    this.cursetfield.len = new DataView(new Uint8Array(rawdata.slice(2, 4)).buffer).getUint16(0);

                    if (this.cursetfield.len > 0) {
                        this.needbytes = this.cursetfield.len;
                        this.curhandlerstep = 1;
                    }
                    else {
                        done = true;
                    }

                    break;
                case 1:
                    this.cursetfield.data = [];

                    this.cursetfield.data = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    done = true;
                    break;
            }

            if (done) {

                logHexDump("SETFIELDDATA", this.cursetfield.data, "Set Field Data Received...");

                var ms = new Message();
                ms.data = this.cursetfield;
                ms.datatype = "object";
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXSETFIELD";

                ws.send2(JSON.stringify(ms));

                logactivity("cursetfieldcount=" + this.cursetfieldcount);

                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;

            }
        };

        // Items Handler
        this.appxitemshandler = function () {
            logactivity("***** SOCKET_ITEMS *****, step=" + this.curhandlerstep);
            var itemDone = false;
            var allDone = false;

            item = function () {
                this.struct = [];
                this.data = [];
            };

            switch (this.curhandlerstep) {
                case 0: // set up to read an item block from the server
                    this.items = [];
                    this.needbytes = 12;
                    this.curhandlerstep = 1;
                    break;
                case 1: // Trim off item block just read and process it and setup to read data if any
                    var itm = new item();
                    itm.struct = this.rtndata.slice(0, 12);
                    this.rtndata = this.rtndata.slice(12, this.rtndata.length);
                    this.needbytes = new DataView(new Uint8Array(itm.struct.slice(10, 12)).buffer).getUint16(0);
                    this.items.push(itm);

                    if (this.needbytes == 0) {
                        if ((itm.struct[9] & 0x02) == 0x02) {
                            //token field, these bytes contain cachid for token field data
                            this.needbytes = 24;
                            this.curhandlerstep = 3;
                        }
                        else {
                            this.needbytes = 2;
                            this.curhandlerstep = 4;
                        }
                    }
                    else {
                        this.curhandlerstep = 2;
                    }

                    break;
                case 2: // we need to read data and that data has arrived, attach it to the last item processed
                    var itm = this.items[this.items.length - 1];
                    itm.data = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    if ((itm.struct[9] & 0x02) == 0x02) {
                        //token field, these bytes contain cachid for token field data
                        this.needbytes = 24;
                        this.curhandlerstep = 3;
                    }
                    else {
                        this.needbytes = 2;
                        this.curhandlerstep = 4;
                    }
                    break;
                case 3: // trim off and process token block
                    var itm = this.items[this.items.length - 1];
                    itm.tokendata = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    this.needbytes = 2;
                    this.curhandlerstep = 4;
                    break;
                case 4: // Trim off and process the widget data length
                    var itm = this.items[this.items.length - 1];
                    itm.widget_len = new DataView(new Uint8Array(this.rtndata.slice(0, 2)).buffer).getUint16(0);
                    this.rtndata = this.rtndata.slice(2, this.rtndata.length);
                    this.needbytes = itm.widget_len;

                    if (this.needbytes == 0) {
                        itm.widget_data = null;
                        this.curhandlerstep = 6;
                        this.needbytes = 4
                    }
                    else {
                        this.curhandlerstep = 5;
                    }
                    break;
                case 5: // trim off and process the widget data
                    var itm = this.items[this.items.length - 1];
                    itm.widget_data = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    this.curhandlerstep++;
                    this.needbytes = 4;
                    break;
                case 6: // trim off and process extra data length
                    // FIXME: add code here
                    var itm = this.items[this.items.length - 1];
                    var bn = this.needbytes;
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getInt32(0);
                    this.rtndata = this.rtndata.slice(bn, this.rtndata.length);
                    if (this.needbytes > 0) {
                        this.curhandlerstep++;
                    }
                    else {
                        if (this.needbytes < 0) {
                            itm.extrareuse = true;
                        }
                        itemDone = true;
                    }
                    break;
                case 7: // trim off and process extra data
                    // FIXME: add code here
                    var itm = this.items[this.items.length - 1];
                    itm.extradata = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    itemDone = true;
                    break;
            }

            if (itemDone) {
                var itm = this.items[this.items.length - 1];
                if (itm.struct[2] == 0) {
                    allDone = true;
                }
                else {
                    this.needbytes = 12;
                    this.curhandlerstep = 1;
                }
            }

            if (allDone) {
                var items = [];
                var i = 0;

                logactivity("processing " + this.items.length + " items...");
                while (this.items.length > 0) {
                    var item = this.items.shift();
                    var itm = new Item();

                    itm.pos_row = item.struct[0];
                    itm.pos_col = item.struct[1];
                    itm.size_rows = item.struct[2];
                    itm.size_cols = item.struct[3];
                    itm.justification = item.struct[4];
                    itm.digits_left = item.struct[5];
                    itm.digits_right = item.struct[6];
                    itm.options = item.struct[7];
                    itm.type = item.struct[8];
                    itm.special = item.struct[9];

                    // sometimes we need the raw data when low order bits are used to delimit data
                    // like formatted fields and date pickers that don't necessarily have an attached
                    // Widget.  Widgets are processed on client, so they already have rawdata.
                    itm.rawdata = item.data

                    itm.data = ab2str(item.data);

                    if (item.widget_data != null) {
                        itm.widget = item.widget_data;
                    }

                    //check itm.special for tokendata
                    if ((itm.special & 0x02) == 0x02) {
                        //cacheid 8
                        //cacheSig 8
                        //token_group 2
                        //ap 3
                        //ver 2
                        //filler 1
                        itm.tokendata = item.tokendata;
                        itm.token_cacheid = ab2str(itm.tokendata.slice(0, 8)).trim();
                        itm.token_cache_sig = ab2str(itm.tokendata.slice(8, 16)).trim();
                        itm.token_group = itm.tokendata.slice(16, 18);
                        itm.token_app = ab2str(itm.tokendata.slice(18, 21)).trim();
                        itm.token_ap_ver = ab2str(itm.tokendata.slice(21, 23)).trim();
                        itm.token_filler = itm.tokendata.slice(23, 24);
                    }

                    logactivity("item " + i + " row=" + itm.pos_row + " col=" + itm.pos_col + " rows=" + itm.size_rows + " cols=" + itm.size_cols + " data=" + itm.data);
                    if (itm.size_rows > 0)
                        items.push(itm);
                    i++;
                }

                var ms = new Message();
                ms.data = items;
                ms.datatype = "array";
                ms.datacount = items.length;
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXITEMS";
                ws.send2(JSON.stringify(ms));

                // setup to read next essage header
                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }
        };

        // Receive File Handler
        this.appxreceivefilehandler = function () {
            logactivity("***** SOCKET_RECEIVE_FILE *****, step=" + this.curhandlerstep);
            switch (this.curhandlerstep) {
                case 0: // set up to read an item block from the server
                    this.needbytes = 4;
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getUint32(0);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);

                    this.curhandlerstep = 2;
                    break;
                case 2:
                    var bdlen = 0;
                    var myfilerecbytes = [];

                    myfilerecbytes = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    logactivity("my url=" + ab2str(myfilerecbytes));

                    var ms = new Message();
                    ms.data = ab2str(myfilerecbytes).trim();
                    ms.datatype = "object";
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXRECEIVEFILE";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.needbytes = 8;
                    this.curhandlerstep = 0;
                    break;
            }
        }

        // Widget Handler
        this.appxwidgetshandler = function () {
            logactivity("***** SOCKET_WIDGETS *****, step=" + this.curhandlerstep);

            var all_done = false;
            var widgetDone = false;

            switch (this.curhandlerstep) {
                case 0: // set up to read the widget count
                    this.widgets = [];
                    this.needbytes = 4;
                    this.curhandlerstep = 1;
                    break;
                case 1: // trim off the widget count
                    this.widgetCount = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getUint32(0);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    if (this.widgetCount == 0) {
                        all_done = true;
                    }
                    else {
                        this.needbytes = 24;
                        this.curhandlerstep = 2;
                    }

                    logactivity("widgetCount=" + this.widgetCount + " all_done=" + all_done + " needbytes=" + this.needbytes);
                    break;
                case 2: // trim off a widget block
                    var mywidget = new WidgetStructure();
                    mywidget.widget_structure_raw_data = this.rtndata.slice(0, 24);
                    this.rtndata = this.rtndata.slice(24, this.rtndata.length);

                    mywidget.type = mywidget.widget_structure_raw_data[0];
                    mywidget.filler = mywidget.widget_structure_raw_data[1];
                    mywidget.parent = new DataView(new Uint8Array(mywidget.widget_structure_raw_data.slice(2, 4)).buffer).getUint16(0);
                    mywidget.pos_row = new DataView(new Uint8Array(mywidget.widget_structure_raw_data.slice(4, 8)).buffer).getUint32(0);
                    mywidget.pos_col = new DataView(new Uint8Array(mywidget.widget_structure_raw_data.slice(8, 12)).buffer).getUint32(0);
                    mywidget.size_row = new DataView(new Uint8Array(mywidget.widget_structure_raw_data.slice(12, 16)).buffer).getInt32(0);
                    mywidget.size_col = new DataView(new Uint8Array(mywidget.widget_structure_raw_data.slice(16, 20)).buffer).getInt32(0);


                    mywidget.data_length = new DataView(new Uint8Array(mywidget.widget_structure_raw_data.slice(20, 24)).buffer).getUint32(0);
                    this.widgets.push(mywidget);

                    if (mywidget.data_length > 0) {
                        this.needbytes = mywidget.data_length;
                        this.curhandlerstep = 3;
                    }
                    else {
                        this.needbytes = 4;
                        this.curhandlerstep = 4;
                    }

                    logactivity(" type=" + mywidget.type + " parent=" + mywidget.parent + " row=" + mywidget.pos_row + " col=" + mywidget.pos_col + " rows=" + mywidget.size_row + " cols=" + mywidget.size_col + " dataLen=" + mywidget.data_length);
                    break;
                case 3:
                    var mywidget = this.widgets[this.widgets.length - 1];
                    mywidget.widget_data = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    this.needbytes = 4;
                    this.curhandlerstep = 4;

                    logactivity("widgetCount=" + this.widgetCount + " widgets.length=" + this.widgets.length + " done=" + all_done + " widget_data=" + ab2str(mywidget.widget_data));
                    break;
                case 4: // trim off and process extra data length
                    // FIXME: add code here
                    var mywidget = this.widgets[this.widgets.length - 1];
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getInt32(0);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);
                    if (this.needbytes > 0) {
                        this.curhandlerstep++;
                    }
                    else {
                        if (this.needbytes < 0) {
                            mywidget.widget_extrareuse = true;
                        }
                        widgetDone = true;
                    }
                    break;
                case 5: // trim off and process extra data
                    // FIXME: add code here
                    var mywidget = this.widgets[this.widgets.length - 1];
                    mywidget.widget_extradata = ab2str(this.rtndata.slice(0, this.needbytes));
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    widgetDone = true;
                    break;
            }

            if (widgetDone) {
                if (this.widgetCount == this.widgets.length) {
                    all_done = true;
                }
                else {
                    this.needbytes = 24;
                    this.curhandlerstep = 2;
                }
            }

            if (all_done == true) {
                var widgets = [];
                while (this.widgets.length > 0) {
                    var wx = this.widgets.shift();
                    var widget_wn = new DataView(new Uint8Array(wx.widget_structure_raw_data.slice(2, 4)).buffer).getUint16(0);

                    //CS - 10/15/2014 - need size_col and size_row for lines, so I tacked on wx to the widget array
                    //FIXME: may need to just send the widget_wn and wx to keep from sending dup data
                    widgets.push([widget_wn, wx.widget_data, {
                        "widget_extra_data": wx.widget_extradata,
                        "widget_extra_reuse": wx.widget_extrareuse == null ? false : true
                    }, wx]);
                } //end while widgets

                var ms = new Message();
                ms.data = widgets;
                ms.datatype = "array";
                ms.datacount = widgets.length;
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXWIDGETS";
                ws.send2(JSON.stringify(ms));

                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }
        };

        // Screen Handler
        this.appxscreenhandler = function () {
            logactivity("***** SOCKET_UPDATE_SCREEN *****, step=" + this.curhandlerstep);

            switch (this.curhandlerstep) {
                case 0:
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 2)).buffer).getUint16(0);
                    this.rtndata = this.rtndata.slice(2, this.rtndata.length);
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    var rd = {};
                    rd.data = "<div id='appx_main_background' style='float:left;width:70%;'>" + "" + "</div>";
                    rd.type = "HTML";

                    var ms = new Message();
                    ms.data = rd;
                    ms.datatype = "object";
                    ms.datacount = 1;
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXSCREEN";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.needbytes = 8;
                    this.curhandlerstep = 0;
                    break;
            }
        };

        // Attributes Handler
        this.appxattributeshandler = function () {
            logactivity("***** SOCKET_UPDATE_ATTRIBUTES *****, step=" + this.curhandlerstep);

            switch (this.curhandlerstep) {
                case 0:
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 2)).buffer).getUint16(0);
                    this.rtndata = this.rtndata.slice(2, this.rtndata.length);
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    var x = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    var ms = new Message();
                    ms.data = x;
                    ms.datatype = "object";
                    ms.datacount = x.length;
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXATTRIBUTES";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.needbytes = 8;
                    this.curhandlerstep = 0;
                    break;
            }
        };

        // Extended Attributes Handler
        this.appxextraattributeshandler = function () {
            logactivity("***** SOCKET_UPDATE_EXTENDED_ATTRIBUTES *****, step=" + this.curhandlerstep);

            switch (this.curhandlerstep) {
                case 0:
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 2)).buffer).getUint16(0);
                    this.rtndata = this.rtndata.slice(2, this.rtndata.length);
                    this.curhandlerstep = 1;
                    break;
                case 1:
                    var x = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);

                    var ms = new Message();
                    ms.data = x;
                    ms.datatype = "object";
                    ms.datacount = x.length;
                    ms.hasdata = "yes";
                    ms.haserrors = "no";
                    ms.type = "APPXEXTRAATTRIBUTES";
                    ws.send2(JSON.stringify(ms));

                    this.curhandler = 0;
                    this.needbytes = 8;
                    this.curhandlerstep = 0;
                    break;
            }
        };

        // Ping Handler
        this.appxpinghandler = function () {
            logactivity("***** SOCKET_PING *****");

            var ms = new Message();
            ms.data = [];
            ms.datatype = "object";
            ms.datacount = 1;
            ms.hasdata = "no";
            ms.haserrors = "no";
            ms.type = "APPXPING";
            ws.send2(JSON.stringify(ms));

            this.curhandler = 0;
            this.needbytes = 8;
            this.curhandlerstep = 0;
        };

        // Show Handler
        this.appxshowhandler = function () {

            logactivity("***** SOCKET_SHOW *****, step=" + this.curhandlerstep);

            var all_done = false;
            var boxDone = false;
            switch (this.curhandlerstep) {
                case 0:
                    this.current_show = new ShowStructure();

                    this.current_show.rawdata = this.rtndata.slice(0, 56);
                    this.rtndata = this.rtndata.slice(56, this.rtndata.length);

                    this.current_show.termid = this.current_show.rawdata.slice(0, 4);
                    this.current_show.curraction = this.current_show.rawdata.slice(4, 5);
                    this.current_show.filler1 = this.current_show.rawdata.slice(5, 8);
                    this.current_show.keymap = this.current_show.rawdata.slice(8, 12);
                    this.current_show.curserrow = this.current_show.rawdata.slice(12, 16);
                    this.current_show.cursorcol = this.current_show.rawdata.slice(16, 20);
                    this.current_show.timeout = new DataView(new Uint8Array(this.current_show.rawdata.slice(20, 24)).buffer).getUint32(0);
                    this.current_show.charatcursor = this.current_show.rawdata.slice(24, 28);
                    this.current_show.useroption = this.current_show.rawdata.slice(28, 32);
                    this.current_show.rtnstatus = this.current_show.rawdata.slice(32, 36);
                    this.current_show.numboxes = this.current_show.rawdata.slice(36, 40);
                    this.current_show.numwdgts = this.current_show.rawdata.slice(40, 44);
                    this.current_show.altcurserrow = null;
                    this.current_show.altcursorcol = null;
                    this.current_show.altuseroption = null;

                    this.current_show.num_boxes = new DataView(new Uint8Array(this.current_show.numboxes).buffer).getUint32(0);
                    this.current_show.boxes = [];

                    logactivity("timeout=" + this.current_show.timeout + " action=" + this.current_show.curraction + " num_boxes=" + this.current_show.num_boxes);
                    if (this.current_show.num_boxes == 0) {
                        all_done = true;
                    }
                    else {
                        this.needbytes = 22;
                        this.curhandlerstep = 1;
                    }

                    break;
                case 1:
                    var box = new TMBoxStructure();

                    box.begin_row = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getUint32(0);
                    box.begin_column = new DataView(new Uint8Array(this.rtndata.slice(4, 8)).buffer).getUint32(0);
                    box.end_row = new DataView(new Uint8Array(this.rtndata.slice(8, 12)).buffer).getUint32(0);
                    box.end_column = new DataView(new Uint8Array(this.rtndata.slice(12, 16)).buffer).getUint32(0);
                    box.bit_mask = new DataView(new Uint8Array(this.rtndata.slice(16, 20)).buffer).getUint32(0);
                    box.data_length = new DataView(new Uint8Array(this.rtndata.slice(20, 22)).buffer).getUint16(0);

                    this.rtndata = this.rtndata.slice(22, this.rtndata.length);

                    this.current_show.boxes.push(box);
                    logactivity("box.data_len=" + box.data_length + " num_boxes=" + this.current_show.num_boxes + " boxes.length=" + this.current_show.boxes.length);
                    if (box.data_length > 0) {
                        this.needbytes = box.data_length;
                        this.curhandlerstep = 2;
                    }
                    else {
                        this.needbytes = 4;
                        this.curhandlerstep = 3;
                    }
                    break;
                case 2:
                    var box = this.current_show.boxes[this.current_show.boxes.length - 1];
                    box.data = this.rtndata.slice(0, box.data_length);
                    this.rtndata = this.rtndata.slice(box.data_length, this.rtndata.length);
                    this.needbytes = 4;
                    this.curhandlerstep = 3;
                    break;
                case 3: // trim off and process extra data length
                    // FIXME: add code here
                    var box = this.current_show.boxes[this.current_show.boxes.length - 1];
                    this.needbytes = new DataView(new Uint8Array(this.rtndata.slice(0, 4)).buffer).getInt32(0);
                    this.rtndata = this.rtndata.slice(4, this.rtndata.length);
                    if (this.needbytes > 0) {
                        this.curhandlerstep++;
                    }
                    else {
                        if (this.needbytes < 0) {
                            box.extrareuse = true;
                        }
                        boxDone = true;
                    }
                    break;
                case 4: // trim off and process extra data
                    // FIXME: add code here
                    var box = this.current_show.boxes[this.current_show.boxes.length - 1];
                    box.extradata = this.rtndata.slice(0, this.needbytes);
                    this.rtndata = this.rtndata.slice(this.needbytes, this.rtndata.length);
                    boxDone = true;
                    break;
            }

            if (boxDone) {
                if (this.current_show.num_boxes == this.current_show.boxes.length) {
                    all_done = true;
                }
                else {
                    this.needbytes = 22;
                    this.curhandlerstep = 1;
                }
            }

            if (all_done) {

                var ms = new Message();
                ms.data = this.current_show;
                ms.datatype = "object";
                ms.datacount = 1;
                ms.hasdata = "yes";
                ms.haserrors = "no";
                ms.type = "APPXSHOW";
                ws.send2(JSON.stringify(ms));

                this.curhandler = 0;
                this.needbytes = 8;
                this.curhandlerstep = 0;
            }
        };
    } //APPXProcessor. object

// APPX File Object
function FileStructure() {
    var self = this;
    self.data = [];
    self.datalength = null;
    self.filenamelength = null;
    self.filename = null;
    self.currdatachunklength = 0;
    self.currdatachunk = [];
    self.filename = null;
}

// Item Object
function Item() {
    var self = this;
    self.pos_row = null;
    self.pos_col = null;
    self.size_rows = null;
    self.size_cols = null;
    self.justification = null;
    self.digits_left = null;
    self.digits_right = null;
    self.options = null;
    self.type = null;
    self.special = null;
    self.data = null; //String converted using ab4str()
    self.rawdata = null; //raws bytes in cases where data in encoded with low order bytes like formatted fields or date picker with masks
    self.widget = null;
}

function Mesg() {
    var self = this;
    self.group = null;
    self.severity = null;
    self.txtlen = null;
    self.txtval = null;
}

// Java Object Class
function JavaObject() {
    var self = this;
    self.createobjectstructurebytes = null;
    self.prototypelength = null;
    self.prototypedata = null;
    self.objecthandle = null;
    self.argumentcount = null;
    self.argumentstructurebytes = null;
    self.argumentdatatype = null;
    self.argumentdatalength = null;
    self.argumentdata = null;
    self.arguments = [];
}

// APPX Menu Object
function Menu() {
    var self = this;
    self.items = [];
}

// APPX Menu Object
function MenuItem() {
    var self = this;
}

// APPX Message Object
function Message() {
    var self = this;
    self.data = [];
    self.hasdata = null;
    self.haserrors = null;
    self.count;
    self.messagepart = 0;
    self.type = null;
}

function TokenStructure() {
    var self = this;

    this.rawdata = null;

    this.ap = null;
    this.ver = null;
    this.group = null;
    this.cacheid = null;
    this.state = 0;

    this.loctype = 0;
    this.ext = 0;
    this.len = 0;
    this.data = [];

    this.id = 0;
    this.ctx = 0;
    this.type = 0;
    this.filler = null;
    this.items = [];

}

function TokenItem() {
    var self = this;
    self.token_id = null;
    self.token_id_group = null;
    self.token_id_val_len = null;
    self.data = "";
}

function SetFieldStructure() {
    var self = this;
    self.row = null;
    self.col = null;
    self.len = null;
    self.data = "";
}

function ResourceStructure() {
    var self = this;
    this.rawdata = null;
    this.ap = null;
    this.ver = null;
    this.cacheid = null;
    this.state = 0;
    this.loctype = 0;
    this.ext = 0;
    this.len = 0;
    this.data = [];
    this.id = 0;
    this.ctx = 0;
    this.type = 0;
    this.filler = null;
}

function ShowStructure() {
    var self = this;
    self.termid = null;
    self.curraction = null;
    filler1 = null;
    keymap = null;
    curserrow = null;
    cursorcol = null;
    timeout = null;
    charatcursor = null;
    useroption = null;
    rtnstatus = null;
    numboxes = null;
    numwdgts = null;
    altcurserrow = null;
    altcursorcol = null;
    altuseroption = null;
    rawdata = null;
    boxdata = null;
    boxes = [];
    num_boxes = 0;
    processed_boxes = 0;
}

function TMBoxStructure() {
    var self = this;
    this.begin_row = 0;
    this.end_row = 0;
    this.begin_column = 0;
    this.end_column = 0;
    this.bit_mask = 0;
    this.data_length = 0;
    this.data = [];
    this.widget = null;
    this.items = [];
    this.widgets = [];
    this.rowtext = [];
};

// Widget Object//function Widget(boxid, html, widgetparams) {

// WidgetStruction Object
function WidgetStructure() {
    var self = this;
    self.type = null;
    self.filler = null;
    self.parent = null;
    self.pos_row = null;
    self.pos_col = null;
    self.size_rows = null;
    self.size_cols = null;
    self.data_length = null;
    self.widget_structure_raw_data = null;
    self.widget_data = null;
    self.num_widgets = 0;
    self.processed_widgets = 0;
    self.table_data = 0;
    self.table_data_length = 0;

}

// Utilities

function ab2int32(buf) {
    return new DataView(new Uint8Array(buf).buffer).getUint32(0);
}

// Convert buffer to string
function ab2str(buf) {
    var str = decoder.write(new Buffer(buf));
    return str;
}

// Convert buffer to string
function ab3str(buf) {
    return String.fromCharCode.apply(" ", new Uint8Array(buf)) + " <br/>";
}

// Convert buffer to string
function ab4str(array) {
    var string = "";
    for (var i = 0; i < array.length; i++) {
        if (array[i] != 0) {
            string += String.fromCharCode(array[i]);
        }
        else {
            string += " ";
        }
    }
    return string + " ";
}

// Convert buffer to string
function buf2str(buf) {
    var str = '';
    var charcode;
    logactivity("got buffer:  " + JSON.stringify(buf));
    for (var i = 0; i < buf.length; i++) {
        charcode = buf.readUInt8(i);
        if (charcode == 0) {
            str += " ";
        }
        else {
            str += String.fromCharCode(charcode);
        }
    }
    buf = [];
    return str;
}

// Convert hex to ascii
function hex2a(hex) {
    var str = '';
    for (var i = 0; i < hex.length; i += 2) {
        str += String.fromCharCode(parseInt(hex.substr(i, 2), 16));
    }
    return str;
}

// Convert string to buffer
function str2ab(str, length) {
    var buf = new Uint16Array(length); // 2 bytes for each char
    for (var i = 0, strLen = str.length; i < strLen; i++) {
        buf[i] = str.charCodeAt(i);
    }
    return buf;
}

// Convert string to buffer
function str3ab(str, length) {
    var buf = new Uint8Array(length); // 2 bytes for each char
    for (var i = 0, strLen = str.length; i < strLen; i++) {
        buf[i] = str.charCodeAt(i);
    }
    return buf;
}

// Convert string to buffer
function str4ab(str) {
    var buf = new Uint8Array(str.length); // 1 bytes for each char
    for (var i = 0, strLen = str.length; i < strLen; i++) {
        buf[i] = str.charCodeAt(i);
    }
    return buf;
}

function base64Array(bytes) { // arrayBuffer
    var base64    = '';
    var encodings = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
    
    //var bytes         = new Uint8Array(arrayBuffer);
    //var byteLength    = bytes.byteLength;
    var byteLength    = bytes.length;
    var byteRemainder = byteLength % 3;
    var mainLength    = byteLength - byteRemainder;
    
    var a, b, c, d;
    var chunk;
 
    // Main loop deals with bytes in chunks of 3
    for (var i = 0; i < mainLength; i = i + 3) {
	// Combine the three bytes into a single integer
	chunk = (bytes[i] << 16) | (bytes[i + 1] << 8) | bytes[i + 2];
 
	// Use bitmasks to extract 6-bit segments from the triplet
	a = (chunk & 16515072) >> 18; // 16515072 = (2^6 - 1) << 18
	b = (chunk & 258048)   >> 12; // 258048   = (2^6 - 1) << 12
	c = (chunk & 4032)     >>  6; // 4032     = (2^6 - 1) << 6
	d = chunk & 63;               // 63       = 2^6 - 1
	
	// Convert the raw binary segments to the appropriate ASCII encoding
	base64 += encodings[a] + encodings[b] + encodings[c] + encodings[d];
    }
 
    // Deal with the remaining bytes and padding
    if (byteRemainder == 1) {
	chunk = bytes[mainLength];
 
	a = (chunk & 252) >> 2; // 252 = (2^6 - 1) << 2
 
	// Set the 4 least significant bits to zero
	b = (chunk & 3)   << 4; // 3   = 2^2 - 1
 
	base64 += encodings[a] + encodings[b] + '==';
    } else if (byteRemainder == 2) {
	chunk = (bytes[mainLength] << 8) | bytes[mainLength + 1];
 
	a = (chunk & 64512) >> 10; // 64512 = (2^6 - 1) << 10
	b = (chunk & 1008)  >>  4; // 1008  = (2^6 - 1) << 4
 
	// Set the 2 least significant bits to zero
	c = (chunk & 15)    <<  2; // 15    = 2^4 - 1
 
	base64 += encodings[a] + encodings[b] + encodings[c] + '=';
    }
  
    return base64;
}
